import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import time as time
from torch.utils.data import DataLoader, TensorDataset
import numpy as np
from torch import Tensor
import random 
from torchvision import datasets, models
import statistics
import collections
import copy
import math
import requests
import socket
import os
from PIL import Image
import scipy.stats as ss
from scipy.stats import norm
from scipy.stats import median_test
from scipy.stats import rankdata
from scipy import stats

#from google.colab import drive
#drive.mount('/content/gdrive')

transform = transforms.Compose(
    [transforms.ToTensor()
    ])

#========DEFINE THE CNN MODEL=====
starting_softmax_neurons = 50
feature_d = 512
feature_w = 5
feature_h = 5
#feature_d = 1024
#feature_w = 3
#feature_h = 3

feature_extractor_size = feature_d*feature_w*feature_h

class CloudClassifier(nn.Module):
    def __init__(self):
        super(CloudClassifier, self).__init__()
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons) #2 output nodes initially
        self.drop1 = nn.Dropout(0.25)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop1(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
print("device: "+str(device))
cloud_cl = CloudClassifier()
cloud_cl = cloud_cl.to(device)

#FUNCTION TO CONVERT STRING BACK TO TENSOR (LABELS AND THE FEATURE MAPS)
def DecomposeString(Feature,batch_size):
    string_tensor = Feature
    listt = string_tensor.split(",")
    arr = np.array(listt,dtype = float)

    depth = int(arr[0])
    width = int(arr[1])
    height = int(arr[2])

    batch_size = int(arr.shape[0]/(depth*width*height))
    
    labels = []
    for i in range(3,3+batch_size):
        labels.append(arr[i])

    torch_labels = torch.from_numpy(np.array(labels,dtype=int))
    #remove first 3 elements from list
    arr = arr[3+batch_size:]

    torch_tensor = torch.from_numpy(arr)
    torch_tensor = torch_tensor.view(-1,depth,width,height)
    return torch_tensor,torch_labels

#Function to create artificial samples
def SMOTE(img_train, lab_train, total_new_points):
    length_of_dataset = len(img_train)
    number_of_points_to_be_chosen = 2 #distance can be calculated using only 2 points

    number_of_data_created = 0
    new_data_img = 0
    new_data_lab = []
    while(number_of_data_created != total_new_points):
        sequence = [i for i in range(length_of_dataset)]
        point_indices = random.sample(sequence, number_of_points_to_be_chosen) #choose any 2 training images from img_train
        
        p1 = img_train[point_indices[0]]
        p2 = img_train[point_indices[1]]

        new_lab = lab_train[point_indices[0]] #choose any point as it will have same index

        number_of_points = 5 if (total_new_points - number_of_data_created) > 5 else (total_new_points - number_of_data_created)
        for i in range(0,number_of_points):
            lamda = np.random.uniform(0,1)
            new_point = p1 + (lamda*(p2-p1))
            new_point = np.reshape(new_point,(1,feature_w,feature_h,feature_d))
            if(number_of_data_created == 0):
                new_data_img = new_point
                new_data_lab.append(new_lab)
            else:
                new_data_img = np.concatenate((new_data_img, new_point), axis=0)
                new_data_lab.append(new_lab)
            #print("new_data_img_shape: "+str(new_data_img.shape))
            number_of_data_created += 1

    new_data_lab = np.asarray(new_data_lab)
    #new_data_img = np.concatenate((new_data_img, img_train), axis=0)
    #new_data_lab = np.concatenate((new_data_lab, lab_train), axis=0)

    return new_data_img,new_data_lab

    """
    #RANDOM OVERSAMPLING (SIMPLY DUPLICATING SOME OF THE FEATURES)
    randomlist = random.sample(range(0, length_of_dataset), total_new_points)
    random_features = img_train[randomlist]
    random_features_labels = lab_train[randomlist]

    #APPEND THE NEW DATA AND THE RESPECTIVE SAMPLES
    new_features = np.concatenate((img_train, random_features), axis=0)
    new_features_labels = lab_train + random_features_labels
    return new_features, new_features_labels
    #print(randomlist)
    """

def GetTrainingAccuracy(cl,device,Trainloader):
    cl.eval()
    correct = 0
    total = 0
    top = 5
    with torch.no_grad():
        for data in Trainloader:
            images, labels = data

            images = images.clone().detach()
            images = images.view(len(images),feature_d,feature_w,feature_h) #RESHAPE THE IMAGES
            labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

            images=images.to(device)
            labels=labels.to(device)

            outputs = cl.forward(images)

            #COMPUTE TOP-5 accuracy
            _,pred=outputs.topk(top,1,largest=True,sorted=True) #select top 5
            total += labels.size(0)

            for j in range(len(pred)):
                tensor = pred[j]
                true_label = labels[j]
                for k in range(len(tensor)):
                    if(true_label == tensor[k].item()):
                        correct += 1
                        break

    acc = float(correct/total)*100.0
    cl.train()
    return acc


def WeightMessageConstruction(best_temp_model,
                              list_of_index_where_updates_are_needed,
                              list_of_index_where_no_updates_needed,
                              message,
                              post_training_bias_list):

    values = [] #weights which must be updated 
    for i in list_of_index_where_updates_are_needed: #indices of useful weights
        values.append(best_temp_model[i]) #useful weights
    
    #convert useful weights to comma separated string
    weight_message_string = ""
    for i in range(len(values)):
        weight_message_string += str(values[i])
        weight_message_string += ","
    weight_message_string += "i,"

    if(message == "y"): #len(list_of_index_where_updates_are_needed) <= len(list_of_index_where_no_updates_needed)
        #convert indices to comma separated string
        for i in range(len(list_of_index_where_updates_are_needed)):
            weight_message_string += str(list_of_index_where_updates_are_needed[i])
            weight_message_string += ","

        #add the biases to the comma separated string
        weight_message_string += "b,"
        for i in range(len(post_training_bias_list)):
            weight_message_string += str(post_training_bias_list[i])
            weight_message_string += ","
        #weight_message_string = weight_message_string[:-1] #remove the last comma
        weight_message_string += "y!"
    else: #len(list_of_index_where_updates_are_needed) > len(list_of_index_where_no_updates_needed)
        for i in range(len(list_of_index_where_no_updates_needed)):
            weight_message_string += str(list_of_index_where_no_updates_needed[i])
            weight_message_string += ","
        weight_message_string += "b,"
        for i in range(len(post_training_bias_list)):
            weight_message_string += str(post_training_bias_list[i])
            weight_message_string += ","
        #weight_message_string = weight_message_string[:-1] #remove the last comma
        weight_message_string += "n!"

    #print("weight message string: "+str(weight_message_string))
    return weight_message_string


def GetThresholdList(weight_difference_list):
    threshold_list = []
    quantiles = 30
    sorted_list = sorted(weight_difference_list)

    divider_space = int(len(sorted_list)/quantiles)

    quantile_index = 0
    for _ in range(0,quantiles):
        threshold_value = sorted_list[quantile_index]
        if(threshold_value>0):
            threshold_list.append(sorted_list[quantile_index])
        quantile_index += divider_space

    return threshold_list


number_of_weights_list = []
number_of_useful_update_list = []
total_params_list = []
total_iterations_list = []
def WeightExtractor(pre_training_weight_list,
                    post_training_weight_list,
                    weight_difference_list,
                    Trainloader,
                    total_nerons_in_final_layer,
                    cl,
                    device,
                    pre_training_bias_list,
                    post_training_bias_list,
                    no_weight_diff_count,
                    message):
    
    global number_of_weights_list
    global number_of_useful_update_list
    global total_params_list
    global total_iterations_list

    totalweightss = total_nerons_in_final_layer*feature_d*feature_w*feature_h
    number_of_weights_list.append(totalweightss)
    total_iterations = 0
    print("\nJUICER ALGORITHM STARTS\n")
    #Get accuracy of the trained post model (training accuracy)
    #original_accuracy = GetTrainingAccuracy(cl,device,Trainloader)
    #print("original trianing accuracy: "+str(original_accuracy))

    threshold_list = GetThresholdList(weight_difference_list)
    print("threshold list: "+str(threshold_list))

    original_cl = copy.deepcopy(cl) #original_cl will now hold the original copy always
    #ori_post_training_weight_list = copy.deepcopy(post_training_weight_list) #ori_post_training_weight_list will now hold the original copy always


    temp_models_list = []
    max_threshold = max(threshold_list)
    print("max threshold is: "+str(max_threshold))
    #original_cl = copy.deepcopy(cl) #original_cl will now hold the original copy always
    #ori_post_training_weight_list = copy.deepcopy(post_training_weight_list) #ori_post_training_weight_list will now hold the original copy always
    temp_model = []
    useful_count = 0
    for j in range(0,len(weight_difference_list)):
        absolute_weight_diff = abs(post_training_weight_list[j] - pre_training_weight_list[j])
        if(absolute_weight_diff>max_threshold): #useful weights if greater than the max threshold
            temp_model.append(post_training_weight_list[j]) #post training weight is useful
            useful_count += 1
        else: #not useful weights
            temp_model.append(pre_training_weight_list[j]) #post training weight is good enough

    #temp_models_list.append(temp_model)
    #Calculate training accuracy here and save in a list
    cl_weight_matrix = torch.from_numpy(np.asarray(temp_model)).float().view(total_nerons_in_final_layer,feature_d*feature_w*feature_h)
    original_cl.softmax_layer.weight.data.copy_(torch.tensor(cl_weight_matrix, requires_grad=True))
    #max_threshold_train_accuracy = GetTrainingAccuracy(cl,device,Trainloader)
    max_threshold_train_accuracy = GetTrainingAccuracy(original_cl,device,Trainloader)
    max_threshold_useful_count = useful_count
    max_threshold_temp_model = temp_model
    print("max threshold max_threshold_train_accuracy: "+str(max_threshold_train_accuracy))
    print("max threshold max_threshold_useful_weights: "+str(max_threshold_useful_count))
    total_iterations += 1



    LList_of_index_where_no_updates_needed = []
    LList_of_index_where_updates_are_needed = []
    temp_models_list = []
    train_accuracy_list = []
    useful_count_list = []
    base_accuracy = 0
    acceptable_training_loss = 0.25
    for i in range(0,len(threshold_list)): #Check for all thresholds

        threshold_value = threshold_list[i]
        original_cl = copy.deepcopy(cl) #original_cl will now hold the original copy always
        ori_post_training_weight_list = copy.deepcopy(post_training_weight_list) #ori_post_training_weight_list will now hold the original copy always
        
        list_of_index_where_no_updates_needed = []
        list_of_index_where_updates_are_needed = []
        temp_model = []
        useful_count = 0
        for j in range(0,len(weight_difference_list)):
            absolute_weight_diff = abs(post_training_weight_list[j] - pre_training_weight_list[j])
            if(absolute_weight_diff>(threshold_value)): #useful weights
                temp_model.append(post_training_weight_list[j]) #post training weight is useful
                list_of_index_where_updates_are_needed.append(j) #useful index
                useful_count += 1 #number of useful weights
            else: #not useful weights
                temp_model.append(pre_training_weight_list[j]) #post training weight is good enough
                list_of_index_where_no_updates_needed.append(j) #post training weight index is good enough
        
        temp_models_list.append(temp_model) #list that contains all the sets of useful weights for each threshold!
        useful_count_list.append(useful_count) #Number of useful weights for each threshold

        #Calculate training accuracy here and save in a list
        cl_weight_matrix = torch.from_numpy(np.asarray(temp_model)).float().view(total_nerons_in_final_layer,feature_d*feature_w*feature_h)
        original_cl.softmax_layer.weight.data.copy_(torch.tensor(cl_weight_matrix, requires_grad=True))

        total_iterations += 1
        train_accuracy = GetTrainingAccuracy(original_cl,device,Trainloader) #Training accuracy of temporary model for each threshold
        train_accuracy_list.append(train_accuracy) #training accuracy of each threshold
        
        if(i==0): #compare accuracy with max threshold accuracy at the very first iteration!
            base_accuracy = train_accuracy #accuracy at the first threshold
            print("BASE accuracy: "+str(base_accuracy))
            best_accuracy = max_threshold_train_accuracy
            accuracy_difference = abs(base_accuracy - best_accuracy)
            print("first attempt checking....")
            if(accuracy_difference<=acceptable_training_loss):
                #best_index = len(threshold_list)-1
                print("first attempt clear, max_threshold_useful_count: "+str(max_threshold_useful_count))
                number_of_useful_update_list.append(max_threshold_useful_count)
                best_temp_model = max_threshold_temp_model

                LList_of_index_where_no_updates_needed = list_of_index_where_no_updates_needed
                LList_of_index_where_updates_are_needed = list_of_index_where_updates_are_needed
                break
            else:
                print("max threshold not the right answer so proceed")
                best_accuracy = train_accuracy
                accuracy_difference = abs(base_accuracy - best_accuracy)
                if(accuracy_difference>=acceptable_training_loss):
                    print("i=0 best weights found")
                    best_temp_model = temp_models_list[i]
                    number_of_useful_update_list.append(useful_count)

                    LList_of_index_where_no_updates_needed = list_of_index_where_no_updates_needed
                    LList_of_index_where_updates_are_needed = list_of_index_where_updates_are_needed
                    break
        else:
            best_accuracy = train_accuracy #accuracy at the current threshold
            print("training accuracy: "+str(train_accuracy)+", useful count: "+str(useful_count))
            accuracy_difference = abs(base_accuracy - best_accuracy)
            print("i="+str(i)+", weight checking")
            if(accuracy_difference>=acceptable_training_loss): #acceptable loss exceeded!
                print("found the answer!, useful weight count: "+str(useful_count_list[i-1])) #loss exceeded so choose previous
                best_temp_model = temp_models_list[i-1] #loss exceeded so choose previous model
                number_of_useful_update_list.append(useful_count_list[i-1]) #List that contains the useful weights for each threshold

                LList_of_index_where_no_updates_needed = list_of_index_where_no_updates_needed
                LList_of_index_where_updates_are_needed = list_of_index_where_updates_are_needed
                break

    total_iterations_list.append(total_iterations) #Number of iterations needed to find useful weights for each threshold
    
    weight_message_string = ""
    list_of_index_where_no_updates_needed = LList_of_index_where_no_updates_needed
    list_of_index_where_updates_are_needed = LList_of_index_where_updates_are_needed
    
    #========THIS PART RUNS AFTER WEIGHT EXTRACTION PROCESS IS FULLY COMPLETE===============
    #list_of_index_where_updates_are_needed = []
    #for i in range(0,len(best_temp_model)): #for all weights in the best temp model
    #    if(i not in list_of_index_where_no_updates_needed):
    #        list_of_index_where_updates_are_needed.append(i)

    #cl = original_cl
    #print("len(list_of_index_where_updates_are_needed)= "+str(len(list_of_index_where_updates_are_needed)))
    #print("len(list_of_index_where_no_updates_needed)= "+str(len(list_of_index_where_no_updates_needed)))

    #number_of_no_update_list.append(len(list_of_index_where_no_updates_needed))
    #total_params_list.append(len(ori_post_training_weight_list))

    #See which index list is bigger, list_of_index_where_updates_are_needed or list_of_index_where_no_updates_needed
    if(len(list_of_index_where_updates_are_needed) <= len(list_of_index_where_no_updates_needed)):
        #if this is the case then send values at indices list_of_index_where_updates_are_needed to the edge along with respective weights
        weight_message_string = WeightMessageConstruction(best_temp_model,list_of_index_where_updates_are_needed,list_of_index_where_no_updates_needed,"y",post_training_bias_list)
    else:
        weight_message_string = WeightMessageConstruction(best_temp_model,list_of_index_where_updates_are_needed,list_of_index_where_no_updates_needed,"n",post_training_bias_list)
    
    print("JUICER ALGORITHM END")
    
    return weight_message_string


test_features_first_time = True
appending_message = False
training_features = 0
training_features_labels = 0
training_features_so_far = 0
training_features_labels_so_far = 0

testing_features = 0
testing_features_labels = 0
testing_features_so_far = 0
testing_features_labels_so_far = 0

#Variables for SMOTE
useful_images_list = []
useful_labels_list = []
length_of_dataset_after_filtration = []
acc_list = []

ori_weight_message_string_length_list = []

training_time_list = []
weight_message_string_length = []
classification_time_list = []
append_time_list = []

incremental_training_accuracy_list = []
smote_time_list = []

Exemplar_features = 0
Exemplar_labels = 0
first_time = 0
incremental_i = 0
cloud_total_nerons_in_final_layer = starting_softmax_neurons
#cloud_temp_cl = CloudTempClassifier()

def Cloud(Feature):
    global test_features_first_time
    global appending_message
    global training_features
    global training_features_labels
    global training_features_so_far
    global training_features_labels_so_far
    global testing_features
    global testing_features_labels
    global testing_features_so_far
    global testing_features_labels_so_far
    global ori_weight_message_string_length_list

    global training_time_list
    global weight_message_string_length
    global classification_time_list
    global append_time_list

    #Variables for SMOTE
    global useful_images_list
    global useful_labels_list
    global length_of_dataset_after_filtration
    global smote_time_list

    global Exemplar_features
    global Exemplar_labels
    global first_time
    global incremental_i
    global acc_list
    global cloud_total_nerons_in_final_layer
    global cloud_cl

    global incremental_training_accuracy_list

    batch_size = 16

    cloud_criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
    cloud_special_criterion = nn.CrossEntropyLoss(reduction='none')
    cloud_learning_rate = 0.0015
    cloud_training_epochs = 25

    #cloud_learning_rate = 0.002
    #cloud_training_epochs = 30
    cloud_optimizer = optim.Adam(cloud_cl.parameters(), lr=cloud_learning_rate) #ADAM OPTIMIZER

    Feature = Feature.replace(" ","") #remove any spaces
    task_character = Feature[-2] #Check the 2nd last character and decide what the program should do now
    print("task_character: "+str(task_character)+"\n")

    if(task_character == 't'): #train the model

        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k') #every feature map is separated by 'k'

        Trainloader = 0
        print("start training \n")
        #All features received from the IoT edge device
        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)

            if(i==0):
                training_features = inputs
                training_features_labels = labels
            else:
                training_features = torch.cat([training_features,inputs],dim=0)
                training_features_labels = torch.cat([training_features_labels,labels],dim=0)
        
        training_features = training_features.float()
        training_features_labels = training_features_labels.float()

        #Joing all the tensors that were received due to RAM overload on IoT edge device
        if(appending_message == True): #This means IoT device has sent a message due to RAM overload
            training_features = torch.cat([training_features,training_features_so_far],dim=0)
            training_features_labels = torch.cat([training_features_labels,training_features_labels_so_far],dim=0)
            appending_message = False
        
        print("training_features before training: "+str(training_features.shape))
        training_features = np.reshape(training_features,(len(training_features),feature_w,feature_h,feature_d))

        #Exemplar loader
        exemplardataset = TensorDataset( Tensor(training_features), Tensor(training_features_labels) )
        Exemplarloader = DataLoader(exemplardataset, batch_size= batch_size, shuffle=False)

        if(first_time==0):

            length_of_dataset_after_filtration.append(len(training_features)) #LENGTH OF SAMPLES IN THE RESPECTIVE TRAINING ROUND
            useful_images_list.append(training_features)
            useful_labels_list.append(training_features_labels)
            
            traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
            Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
        else:
            #CARRY OUT SMOTE OVER HERE
            if(len(training_features)<=length_of_dataset_after_filtration[incremental_i-1]):
                print("\ncurrent incremental training round in minority ")
                if(len(training_features)<length_of_dataset_after_filtration[incremental_i-1]):
                    total_new_points = length_of_dataset_after_filtration[incremental_i-1] - len(training_features)
                    smote_tic = time.time()
                    remaining_images,remaining_labels = SMOTE(training_features, training_features_labels, total_new_points)
                    smote_toc = time.time()
                    smote_time = smote_toc - smote_tic
                    smote_time_list.append(smote_time)
                    #remaining_images = np.reshape(remaining_images,(total_new_points,feature_extractor_size,w,h))
                    remaining_images = np.reshape(remaining_images,(total_new_points,feature_w,feature_h,feature_d))

                    print("training_features shape: "+str(training_features.shape))
                    print("remaining_images shape: "+str(remaining_images.shape))
                    training_features = np.concatenate((training_features,remaining_images),axis=0)
                    training_features_labels = np.concatenate((training_features_labels,remaining_labels),axis=0)
                    traindataset = TensorDataset( Tensor(training_features), Tensor(training_features_labels) )
                    Trainloader_trail = DataLoader(traindataset, batch_size= batch_size, shuffle=False)
                    #get initial losses
                    traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
                    Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
                    #length_of_dataset_after_filtration[incremental_i-1] = len(training_features)

                    print("length of new images: "+str(len(training_features)))
                    length_of_dataset_after_filtration.append(len(training_features))
                    useful_images_list.append(training_features)
                    useful_labels_list.append(training_features_labels)
                else:
                    print("same number of images as last time")
                    length_of_dataset_after_filtration.append(len(training_features))
                    useful_images_list.append(training_features)
                    useful_labels_list.append(training_features_labels)

                    traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
                    Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
            elif(len(training_features)>length_of_dataset_after_filtration[incremental_i-1]):
                n_images_total = 0
                n_labels_total = 0
                print("length of useful_images_list: "+str(len(useful_images_list)))
                for kk in range(0,len(useful_images_list)):
                    total_new_points = len(training_features) - len(useful_images_list[kk])
                    smote_tic = time.time()
                    n_images,n_labels = SMOTE(useful_images_list[kk], useful_labels_list[kk], total_new_points)
                    smote_toc = time.time()
                    smote_time = smote_toc - smote_tic
                    smote_time_list.append(smote_time)
                    n_images = np.reshape(n_images,(total_new_points,feature_w,feature_h,feature_d))
                    print("useful_images_list[kk] shape: "+str(useful_images_list[kk].shape))
                    print("n_images(new artifical images) shape: "+str(n_images.shape))
                    useful_images_list[kk] = np.concatenate((useful_images_list[kk],n_images),axis=0)
                    useful_labels_list[kk] = np.concatenate((useful_labels_list[kk],n_labels),axis=0)
                    length_of_dataset_after_filtration[kk] = len(useful_images_list[kk])

                    if(kk==0):
                        n_images_total = n_images
                        n_labels_total = n_labels
                    else:
                        n_images_total = np.concatenate((n_images_total,n_images),axis=0)
                        n_labels_total = np.concatenate((n_labels_total,n_labels),axis=0)
                
                print("n_images_total(TOTAL artificial samples) shape: "+str(n_images_total.shape))
                print("current images_train shape: "+str(training_features.shape))
                n_images_total = np.concatenate((n_images_total,training_features),axis=0)
                n_labels_total = np.concatenate((n_labels_total,training_features_labels),axis=0)
                print("n_images_total(TOTAL appended samples) shape: "+str(n_images_total.shape))

                traindataset_new = TensorDataset( Tensor(n_images_total), Tensor(n_labels_total) )
                Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
                

                length_of_dataset_after_filtration.append(len(training_features))
                useful_images_list.append(training_features)
                useful_labels_list.append(training_features_labels)

            current_features_train = 0
            current_features_label = 0

            #Get features of the current class
            for i, data in enumerate(Trainloader, 0):
                inputs, labels = data
                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                Features = inputs
                #Features = Features.to(device)
                #labels = labels.to(device)

                if(i==0):
                    current_features_train = Features
                    current_features_label = labels
                else:
                    current_features_train = torch.cat([current_features_train,Features],dim=0)
                    current_features_label = torch.cat([current_features_label,labels],dim=0)

            #APPEND THE CURRENT SAMPLES FEATURES WITH THE EXEMPLARS
            appended_dataset_image = torch.cat([Exemplar_features,current_features_train],dim=0)
            appended_dataset_label = torch.cat([Exemplar_labels,current_features_label],dim=0)

            print("appended data shape is: "+str(appended_dataset_image.shape)+" and appended label shape is: "+str(appended_dataset_label.shape)+"\n")
            dataset = TensorDataset( appended_dataset_image, appended_dataset_label )
            Trainloader = DataLoader(dataset, batch_size= batch_size, shuffle=True)

        print("data loader prepared \n")

        #pre_training_cl = copy.deepcopy(cloud_cl)
        #pre_training_weight_list = pre_training_cl.softmax_layer.weight.view(-1,feature_extractor_size*cloud_total_nerons_in_final_layer).cpu().detach().numpy().tolist()[0]
        #pre_training_bias_list = pre_training_cl.softmax_layer.bias.view(-1,cloud_total_nerons_in_final_layer).cpu().detach().numpy().tolist()[0]

        print("before training, cloud_cl is: "+str(cloud_cl))
        training_time_tic = time.time()
        #Train all the features on the cloud now
        for epoch in range(0,cloud_training_epochs):
            for i, data in enumerate(Trainloader, 0):

                inputs, labels = data

                #Pass the data now to the neural network
                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                outputs = cloud_cl.forward(inputs)
                loss = cloud_criterion(outputs, labels)
                        
                cloud_optimizer.zero_grad()
                loss.backward()
                cloud_optimizer.step()

        training_time_toc = time.time()
        training_time = (training_time_toc-training_time_tic)
        training_time_list.append(training_time)

        #incremental_training_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)
        #incremental_training_accuracy_list.append(incremental_training_accuracy)

        print("done training \n")
        training_features = 0
        training_features_labels = 0
        first_time = 1


        #print("Exemplar loading, incremental_i value: "+str(incremental_i))
        for i, data in enumerate(Exemplarloader, 0):
            inputs, labels = data

            inputs = inputs.clone().detach()
            inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
            labels = labels.long()

            Features = inputs
            #print("Features shape: "+str(Features.shape))

            #print("i = "+str(i)+", incremental_i: "+str(incremental_i))
            #Features = inputs #RAW IMAGES ARE EXEMPLARS 
            if(i==0 and incremental_i==0):
                with torch.no_grad():
                    Exemplar_features = Features
                    Exemplar_labels = labels
                    #print("NN at iter: "+str(i)+" exemplar shape is: "+str(Exemplar_features.shape)+" and label shape is: "+str(Exemplar_labels.shape)+"\n")
            else:
                with torch.no_grad():
                    #print("Exemplar_features shape: "+str(Exemplar_features.shape))
                    Exemplar_features = torch.cat([Exemplar_features,Features],dim=0)
                    Exemplar_labels = torch.cat([Exemplar_labels,labels],dim=0)
                    #print("at iter: "+str(i)+" exemplar shape is: "+str(Exemplar_features.shape)+" and label shape is: "+str(Exemplar_labels.shape)+"\n")

        incremental_i += 1 #INCREMENTAL TRAINING ROUND COUNTER
        #appending_message = False

        #paramter_string_size = len(weight_message_string)*1e-6
        #weight_message_string_length.append(paramter_string_size)

        #print("weight message string: "+str(weight_message_string))

    elif(task_character == 'h'):
        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k')

        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            testing_features = inputs
            testing_features_labels = labels

            testing_features = testing_features.float()
            testing_features_labels = testing_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and test_features_first_time == True): #The very first time we are processing an append message
                test_features_first_time = False
                testing_features_so_far = testing_features
                testing_features_labels_so_far = testing_features_labels
            else:
                testing_features_so_far = torch.cat([testing_features_so_far,testing_features],dim=0)
                testing_features_labels_so_far = torch.cat([testing_features_labels_so_far,testing_features_labels],dim=0)

        print("cloud testing append, testing_features_so_far shape: "+str(testing_features_so_far.shape))

    elif(task_character == 'c'): #time for classification (overall accuracy)
        #optimizer = optim.Adam(cl.parameters(), lr=learning_rate, weight_decay=0.00001) #ADAM OPTIMIZER
        #print("classification start now, the model is: "+str(cloud_cl))
        classification_time_tic = time.time()
        cloud_cl.eval() #put the model is evaluation mode

        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k')


        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            testing_features = inputs
            testing_features_labels = labels

            testing_features = testing_features.float()
            testing_features_labels = testing_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and test_features_first_time == True): #The very first time we are processing an append message
                test_features_first_time = False
                testing_features_so_far = testing_features
                testing_features_labels_so_far = testing_features_labels
            else:
                testing_features_so_far = torch.cat([testing_features_so_far,testing_features],dim=0)
                testing_features_labels_so_far = torch.cat([testing_features_labels_so_far,testing_features_labels],dim=0)

        print("classification testing_features_so_far shape is: "+str(testing_features_so_far.shape))
        testdatasett = TensorDataset( testing_features_so_far, testing_features_labels_so_far )
        Testloaderr = DataLoader(testdatasett, batch_size= batch_size, shuffle=False)

        correct = 0
        total = 0
        top = 5
        print(cloud_cl)
        with torch.no_grad():
            for data in Testloaderr:
                images, labels = data

                images = images.clone().detach()
                images = images.view(len(images),feature_d,feature_w,feature_h) #RESHAPE THE IMAGES
                labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

                images=images.to(device)
                labels=labels.to(device)

                #outputs = squeezenet.features(images)
                outputs = cloud_cl.forward(images)

                #COMPUTE TOP-5 accuracy
                _,pred=outputs.topk(top,1,largest=True,sorted=True) #select top 5
                total += labels.size(0)

                for j in range(len(pred)):
                    tensor = pred[j]
                    true_label = labels[j]
                    for k in range(len(tensor)):
                        if(true_label == tensor[k].item()):
                            correct += 1
                            break
        if(total>0):
            acc = float(correct/total)*100.0
            print("correct is: "+str(correct)+"\n")
            print("total is: "+str(total)+"\n")
            print("accuracy is: "+str(acc)+"\n")
            acc_list.append(acc)

        cloud_cl.train()

        #ADD NEURONS TO THE FINAL SOFTMAX CLASSIFICATION LAYER
        print("modifying the cloud model now \n")
        
        #Add nerons to the fina layer and set the new random weights
        current_weights = cloud_cl.softmax_layer.weight.data
        current_bias = cloud_cl.softmax_layer.bias.data

        cloud_total_nerons_in_final_layer = cloud_total_nerons_in_final_layer + starting_softmax_neurons
        cloud_cl.softmax_layer = nn.Linear(feature_extractor_size,cloud_total_nerons_in_final_layer)

        print("cloud_total_nerons_in_final_layer: "+str(cloud_total_nerons_in_final_layer))

        new_weights = -0.0001 * np.random.random_sample((starting_softmax_neurons, feature_extractor_size)) + 0.0001 #RANDOMLY INITIALIZED WEIGHTS FOR THE NEW TASKS
        new_weights = torch.from_numpy(new_weights).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
        new_weights = new_weights.to(device)
        new_weights = torch.cat([current_weights,new_weights],dim=0) #Concatenate the new initialized weights with the old weights for old task
        cloud_cl.softmax_layer.weight.data.copy_(torch.tensor(new_weights, requires_grad=True))

        new_biases = np.zeros(starting_softmax_neurons)
        new_biases = torch.from_numpy(new_biases).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
        new_biases = new_biases.to(device)
        new_biases = torch.cat([current_bias,new_biases],dim=0)
        cloud_cl.softmax_layer.bias.data.copy_(torch.tensor(new_biases, requires_grad=True))

        #After adding neurons to the final layer, the network needs to be pushed to the GPU again
        cloud_cl = cloud_cl.to(device)
        #print("cloud_cl after appending is: "+str(cloud_cl))

        cloud_optimizer = optim.Adam(cloud_cl.parameters(), lr=cloud_learning_rate) #UPDATE OPTIMIZER

        classification_time_toc = time.time()
        classification_time = (classification_time_toc - classification_time_tic)
        classification_time_list.append(classification_time)

        print("classification done\n")
    
    elif(task_character == 'a'): #just append the tensors and send back acknowledgement back to IoT edge device

        append_tic = time.time()
        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k') #every feature map is separated by 'k'

        print("start appending \n")
        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            training_features = inputs
            training_features_labels = labels

            training_features = training_features.float()
            training_features_labels = training_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and appending_message == False): #The very first time we are processing an append message
                training_features_so_far = training_features
                training_features_labels_so_far = training_features_labels
            else:
                training_features_so_far = torch.cat([training_features_so_far,training_features],dim=0)
                training_features_labels_so_far = torch.cat([training_features_labels_so_far,training_features_labels],dim=0)

        print("training_features_so_far shape: "+str(training_features_so_far.shape))
        print("done appending")
        appending_message = True
        append_toc = time.time()
        append_time_list.append(append_toc-append_tic)

    elif(task_character == 'd'): #means process done
        print("============================CLOUD RESULTS=====================================")
        print("number of classes being learnt at a time: "+str(starting_softmax_neurons))
        print("overall training_time_list: "+str(training_time_list))
        print("TOTAL TRAINING TIME (SECONDS): "+str(sum(training_time_list)))
        print("Testing acc_list: "+str(acc_list))
        print("smote_time_list: "+str(smote_time_list))
        print("TOTAL smote_time_list: "+str(sum(smote_time_list)))
        


#======================================CLOUD CODE END HERE========================================
















#======================================EDGE CODE DOWN HERE========================================
class EdgeClassifier(nn.Module):
    def __init__(self):
        super(EdgeClassifier, self).__init__()
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons) #2 output nodes initially
        self.drop1 = nn.Dropout(0.25)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop1(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

class EdgeTempClassifier(nn.Module):
    def __init__(self):
        super(EdgeTempClassifier, self).__init__()
        self.drop2 = nn.Dropout(0.25)
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop2(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

def TorchNormalizedTrainingData():
    print("Get sorted training data\n")
    img_list = []
    lab_list = []
    with torch.no_grad():
        for data in Trainloader_fake:
            images, labels = data
            for j in range(0,len(images)):
              img_list.append(np.array(images[j]))
              lab_list.append(int(labels[j]))

    img_list = np.array(img_list)    
    lab_list = np.array(lab_list)
    #Sort training dataset wrt to labels

    sorted_X_train = [] #this will contains training images sorted based on the label number
    sorted_Y_train = [] #label
    start_index = 0
    end_index = 100
    for j in range(start_index,end_index):
      searchval = j
      all_positions_of_label = np.where(lab_list == searchval)[0] #Find all indexes that contains `searchval`
      for j in range(0,len(all_positions_of_label)):
          image = img_list[all_positions_of_label[j]]
          sorted_X_train.append(image)
          sorted_Y_train.append(searchval)
    
    sorted_X_train = np.array(sorted_X_train)
    sorted_X_train = sorted_X_train.reshape(50000, 32, 32, 3)
    #sorted_Y_categorical_train = np_utils.to_categorical(sorted_Y_train,end_index-start_index)

    return sorted_X_train,sorted_Y_train#,sorted_Y_categorical_train

def TorchNormalizedTestingData():
    img_list = []
    lab_list = []
    with torch.no_grad():
      for data in Testloader:
        images, labels = data
        for j in range(0,len(images)):
          img_list.append(np.array(images[j]))
          lab_list.append(int(labels[j]))

    img_list = np.array(img_list)    
    lab_list = np.array(lab_list)
    #Sort training dataset wrt to labels

    sorted_X_test = [] #this will contains training images sorted based on the label number
    sorted_Y_test = [] #label
    start_index = 0
    end_index = 100
    for j in range(start_index,end_index):
      searchval = j
      all_positions_of_label = np.where(lab_list == searchval)[0] #Find all indexes that contains `searchval`
      for j in range(0,len(all_positions_of_label)):
        image = img_list[all_positions_of_label[j]]
        sorted_X_test.append(image)
        sorted_Y_test.append(searchval)
    sorted_X_test = np.array(sorted_X_test)
    sorted_X_test = sorted_X_test.reshape(10000, 32, 32, 3)
    #sorted_Y_categorical_test = np_utils.to_categorical(sorted_Y_test,end_index-start_index)
    
    return sorted_X_test,sorted_Y_test#,sorted_Y_categorical_test

#ANY DATA RECEIVED BY THE CLOUD WILL BE TAKEN CARE OF BY THIS FUNCTION
def ReceiveDataFromCloud(client):
    recevied_data = " "
    received_result = ""
    print("receving start")
    while recevied_data[-1]!="!":
        data = client.recv(10000).decode("utf-8")
        recevied_data += data
        received_result += data
    #print("data received is: "+str(received_result)+"\n")
    #pos = (int)(received_result[0])
    client.close()
    return received_result
    #return 0

#FUNCTION FOR CONNECTED TO GOOGLE CLOUD VM AND TRANSMITTING DATA
def SendDataToCloud(Feature):

    #Connect to the cloud and send the data
    client = socket.socket(socket.AF_INET,socket.SOCK_STREAM)#AF_NET = ipv4, SOCK_STREAM = tcp oriented
    #ip = '34.80.68.8'
    ip = '34.80.253.190'
    #ip = '104.155.209.67'
    port = 3389
    address = (ip,port) #address of the server
    client.connect(address)
    print("sending to the cloud \n")
    
    #Send the string to the cloud
    client.send(Feature.encode())
    print("sent to the cloud \n")

    #Receive response from the server and only then continue
    received_string = ReceiveDataFromCloud(client)
    print("recvd from the cloud \n")
    return received_string

#FUNCTION FOR SENDING TEST DATASET TO THE CLOUD FOR COMPUTING CLASSIFICATION ACCURACY ON TEST DATASET
def CalculateModelAccuracy(squeezenet,device,Testloader,batch_size):
    
    overall_string_tensor = ""
    with torch.no_grad():
        for data in Testloader:
            images, labels = data

            images = images.clone().detach()
            images = images.view(len(images),3,32,32) #RESHAPE THE IMAGES
            labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

            images=images.to(device)
            labels=labels.to(device)

            outputs = squeezenet.features(images)

            #Convert the tensor to string
            string_tensor = TensorToString(labels,outputs)
            overall_string_tensor = overall_string_tensor + string_tensor + "k"

        #Send the string to the cloud
        overall_string_tensor = overall_string_tensor[:-1] + "c!"
        Cloud(overall_string_tensor)
    

def GetLosses(Trainloader_trail,device,cls):

    special_criterion = nn.CrossEntropyLoss(reduction='none')

    initial_loss_list = []

    image_loss_dictionary = collections.defaultdict(list)
    label_loss_dictionary = collections.defaultdict(list)

    for _, data in enumerate(Trainloader_trail, 0):
        inputs, labels = data

        inputs = inputs.clone().detach()
        inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
        labels = labels.long()
                
        inputs=inputs.to(device)
        labels=labels.to(device)

        with torch.no_grad():
            outputs = cls.forward(inputs)

            #compute softmax here
            """
            softmax_function = nn.Softmax(dim=1)
            softmax_output = softmax_function(outputs)
            loss = []
            for j in range(len(softmax_output)):
                loss.append(softmax_output[j][labels[j]])
            """
            loss = special_criterion(outputs, labels)
            for j in range(0,len(loss)):
                initial_loss_list.append(loss[j].item())
                image_loss_dictionary[loss[j].item()].append(inputs[j])
                label_loss_dictionary[loss[j].item()].append(labels[j].cpu().numpy())

    return initial_loss_list, image_loss_dictionary, label_loss_dictionary

#FUNCTION FOR COMPUTING THE STATISTICS OF THE SORTED LOSS VALUES OF THE TRAINING IMAGES
def GetLossStatistics(sorted_initial_list):
    #Calculate mean
    mean = statistics.mean(sorted_initial_list)

    #Calculate median
    median = statistics.median(sorted_initial_list)

    #Calculate standard deviation
    standard_deviation = statistics.stdev(sorted_initial_list)

    
    mid_range_loss = sorted_initial_list[int(len(sorted_initial_list)/2)]
    C = []
    for i in range(0,len(sorted_initial_list)):
        diff_sqaured = (sorted_initial_list[i]-mid_range_loss)**2
        C.append(diff_sqaured)
    standard_deviation = math.sqrt(sum(C)/len(C))
    
    return mean, median, standard_deviation

def MannWhitneyTest(loss_comparison_list,losses):
    alpha = 0.05

    A = loss_comparison_list
    B = losses
    #C = A + B #Step 1: Concatenate the lists
    C = []
    for i in range(0,len(A)):
      C.append(A[i])
    for i in range(0,len(B)):
      C.append(B[i])
    #C.extend(A)
    #C.extend(B)
    #print("Concatenated: "+str(C))

    ranks = ss.rankdata(C) #Compute the ranks of the concatenrated lists
    #print("ranks: "+str(ranks))

    ranksum_A = sum(ranks[0:len(A)]) #Ranksum of list A
    ranksum_B = sum(ranks[len(A):]) #Ranksum of list B

    #print("ranksum_A: "+str(ranksum_A))
    #print("ranksum_B: "+str(ranksum_B))

    n1 = len(A) #Number of of elements in list A
    n2 = len(B) #Number of of elements in list B

    #print("n1: "+str(n1))
    #print("n2: "+str(n2))

    mean = (n1*(n1 + n2 + 1))/2.0 #Mean
    stddev = math.sqrt((n1*n2*(n1 + n2 + 1))/12.0) #Standard deviation

    #print("mean: "+str(mean))
    #print("stddev: "+str(stddev))

    z = (ranksum_A - mean)/stddev #Z-test
    p_value = norm.cdf(z) #p value

    #print("z: "+str(z))
    #print("p_value: "+str(p_value))

    if(p_value < alpha): #Enough evidence that 2 sets are different
        return 1
    else: #Not enough evidence that 2 sets are different
        return 0

def Kruskal_Wallis_Test(array_list):
    concatenated_list = []
    for i in range(0,len(array_list)):
        concatenated_list += array_list[i]
    ranks = rankdata(concatenated_list)
    n_T = len(concatenated_list)
    R_i = []
    n_i = []
    start = 0
    for i in range(0,len(array_list)):
        end = start + len(array_list[i])
        R_i.append(sum(ranks[start:end]))
        n_i.append(len(array_list[i]))
        start = end

    part_1 = 12/(n_T*(n_T+1))
    part_2 = sum([i / j for i, j in zip([i**2 for i in R_i], n_i)])
    part_3 = 3*(n_T + 1)
    H = (part_1*part_2)-part_3
    #print("H: "+str(H))

    chi_square_value = stats.chi2.ppf(0.95 , len(array_list)-1)
    #print(chi_square_value)
    return H, chi_square_value

#FUNCTION FOR FINDING THE CUTOFF LOSS VALUE AND DISCARDING THE SAMPLES
old_mean_loss = 0
def Cutoff(mean,median,standard_deviation,image_loss_dictionary,label_loss_dictionary,depth,width,height):
    sorted_loss_list = sorted(image_loss_dictionary)
    number_of_samples_to_be_dropped = 0

    """
    cutoff_loss_value = min(sorted_loss_list)
    min_loss_value = min(sorted_loss_list)
    for i in range(0,len(sorted_loss_list)):
        distance = median - sorted_loss_list[i]
        if(distance < standard_deviation):
            cutoff_loss_value = sorted_loss_list[i]
            print("min loss: "+str(min_loss_value)+", distance: "+str(distance)+" , std: "+str(standard_deviation))
            break
    
    #cutoff_loss_value = cutoff_loss_value - (0.25*(cutoff_loss_value-min_loss_value))
    number_of_samples_to_be_dropped = 0
    for i in range(0,len(sorted_loss_list)):
        current_loss = sorted_loss_list[i]
        if(current_loss < cutoff_loss_value):
            number_of_samples_to_be_dropped += 1
        else:
            break
    
    omega = 1.0
    number_of_samples_to_be_dropped = int(omega * number_of_samples_to_be_dropped)
    """

    """
    #WRSTS Sampling
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        base_losses = sorted_loss_list
        val = MannWhitneyTest(base_losses,test_losses)
        if(val == 1):
            print("mann-whitney deviation is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    """

    """
    #MTS Sampling
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        base_losses = sorted_loss_list
        stat, p, med, tbl = median_test(base_losses,test_losses)
        if(p < 0.05):
            print("median deviation is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    """

    
    #KWT Sampling
    base_losses = sorted_loss_list
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        array_list = [test_losses,base_losses]
        #array_list.append(test_losses)
        #array_list.append(base_losses)
        #H,p=stats.mstats.kruskalwallis(array_list)
        H,chi_square_value = Kruskal_Wallis_Test(array_list)
        #print("HH: "+str(H))
        #print("chi_square_value: "+str(chi_square_value))
        #chi_square_value = stats.chi2.ppf(0.95 , len(array_list)-1)
        if(H > chi_square_value): #If true then reject the null hypothesis which means the 2 groups not the same
            print("Kruskal wallis is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    
    #number_of_samples_to_be_dropped = 0
    print("number_of_samples_to_be_dropped: "+str(number_of_samples_to_be_dropped))

    #get all the images and labels above the cutoff value
    number_of_samples_to_selected = 0
    new_images = 0
    new_labels = []
    filtered_original_losses = []

    """
    number_of_samples_to_be_selected = 0
    #randomly select losses
    if(number_of_samples_to_be_dropped > 0):
        number_of_samples_to_be_selected = len(sorted_loss_list) - number_of_samples_to_be_dropped
        print("number_of_samples_to_be_selected: "+str(number_of_samples_to_be_selected))
        randomly_selected_losses = random.sample(sorted_loss_list, k=number_of_samples_to_be_selected)
    else:
        randomly_selected_losses = random.sample(sorted_loss_list, k=len(sorted_loss_list))
        print("number_of_samples_to_be_selected: "+str(len(sorted_loss_list)))
        print("must select all the samples now!!!!! ")
    
    #RANDOM SAMPLING
    
    for i in range(0,len(randomly_selected_losses)):
        key_loss = randomly_selected_losses[i]
        #if(key_loss not in randomly_selected_losses):
        selected_image = image_loss_dictionary[key_loss]
        selected_label = label_loss_dictionary[key_loss]

        #if(len(selected_image)>1):
        #      print("length of selected_image: "+str(len(selected_image)))
        if(number_of_samples_to_selected == 0):
            for j in range(0,len(selected_image)): # aloss key can have duplicates
                if(j==0):
                    new_images = selected_image[j]
                else:
                    new_images = torch.cat([new_images,selected_image[j]],dim=0)
                #print("type of label is: "+str(type(selected_label[j])))
                new_labels.append(selected_label[j])
                filtered_original_losses.append(key_loss)
        else:
            #print("betrne")
            for j in range(0,len(selected_image)): # aloss key can have duplicates
                new_images = torch.cat([new_images,selected_image[j]],dim=0)
                new_labels.append(selected_label[j])
                filtered_original_losses.append(key_loss)
        number_of_samples_to_selected += len(selected_image)
    """
    #ACTIVE LEARNING ENTROPY BASED SAMPLING
    
    cutoff_loss_value = sorted_loss_list[number_of_samples_to_be_dropped]
    for i in range(len(sorted_loss_list)-1,-1,-1):
        #print("i is: "+str(i))
        key_loss = sorted_loss_list[i]
        #print("key loss type is: "+str(type(key_loss))+" and value is: "+str(key_loss))
        if(key_loss>=cutoff_loss_value):
            selected_image = image_loss_dictionary[key_loss]
            selected_label = label_loss_dictionary[key_loss]
            #print("key: "+str(key_loss)+" and value: "+str(data[key_loss]))
            #if(len(selected_image)>1):
              #print("length of selected_image: "+str(len(selected_image)))
            if(number_of_samples_to_selected == 0):
                for j in range(0,len(selected_image)): # aloss key can have duplicates
                    if(j==0):
                        new_images = selected_image[j]
                    else:
                        new_images = torch.cat([new_images,selected_image[j]],dim=0)
                    #print("type of label is: "+str(type(selected_label[j])))
                    new_labels.append(selected_label[j])
                    filtered_original_losses.append(key_loss)
            else:
                #print("betrne")
                for j in range(0,len(selected_image)): # aloss key can have duplicates
                    new_images = torch.cat([new_images,selected_image[j]],dim=0)
                    new_labels.append(selected_label[j])
                    filtered_original_losses.append(key_loss)
            number_of_samples_to_selected += len(selected_image)
        else:
            break
    #"""
    new_images = new_images.reshape((number_of_samples_to_selected,depth,width,height))
    print("new_images shape: "+str(new_images.shape))

    new_labels = torch.from_numpy(np.asarray(new_labels)).float()
    print("new_labels shape: "+str(new_labels.shape))

    #filtered_original_losses = torch.from_numpy(np.asarray(filtered_original_losses)).float()
    #print("filtered_original_losses shape: "+str(filtered_original_losses.shape))
    return new_images,new_labels#,filtered_original_losses

#MAIN FUNCTION FOR PERFORMING DATA SAMPLING
def GetusefulData(images_train,labels_train,batch_size,cls):

    traindataset = TensorDataset( images_train, Tensor(labels_train) )
    Trainloader_trail = DataLoader(traindataset, batch_size= batch_size, shuffle=False)

    #Do single forward pass and get the losses
    #cl.eval()
    initial_loss_list, image_loss_dictionary, label_loss_dictionary = GetLosses(Trainloader_trail,device,cls)
    sorted_initial_list = sorted(initial_loss_list)
    mean, median, standard_deviation = GetLossStatistics(sorted_initial_list)
    #print("mean: "+str(mean))
    #print("median: "+str(median))
    #print("standard_deviation: "+str(standard_deviation))
    new_images, new_labels = Cutoff(mean,median,standard_deviation,image_loss_dictionary,label_loss_dictionary,feature_d,feature_w,feature_h)
    

    rejection_rate = (1.0 - (len(new_images)/len(images_train)))*100.0
    print("rejection rate: "+str(rejection_rate))
    return new_images.cpu().numpy(), new_labels.cpu().numpy()



#FUNCTION TO CONVERT TENSOR TO STRING
def TensorToString(labels,outputs):

    tensor = outputs
    tensor = tensor.view(-1,outputs.shape[1]*outputs.shape[2]*outputs.shape[3]) #flatten the tensor

    string_tensor = ""
    string_tensor = str(outputs.shape[1]) + "," + str(outputs.shape[2]) + "," + str(outputs.shape[3]) + ","
    mini_batch_length = len(labels)
    for i in range(0,mini_batch_length):
        string_tensor = string_tensor + str(labels[i].item()) + ","


    new_string = string_tensor
    for i in range(0,mini_batch_length):
        numpy_tensor = tensor[i].cpu().numpy()
        out_string = ','.join(str(x) for x in numpy_tensor)
        new_string = new_string + out_string + ","
    new_string = new_string[:-1] #remove last comma
    #new_string = new_string + "t!" 

    string_tensor = new_string
    return string_tensor

#FUNCTION TO DECODE WEIGHTS AND BIASES (SEPARATING WEIGHTS & BIASES FROM COMMAS)
def DecodeWeightsAndBiases(message):
    task_character = message[-2]

    bias_string = message[message.find("b")+2:message.find(task_character)-1]
    bias_list = list(map(float, bias_string.split(",")))

    weight_string = message[0:message.find("i")-1]
    weight_list = list(map(float, weight_string.split(","))) 

    indices_string = message[message.find("i")+2:message.find("b")-1]
    indices_list = indices_string.split(",")

    return weight_list, bias_list, indices_list, task_character

def DataSampling(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl, edge_temp_cl,number_of_classes_to_be_added_ever_increment,incremental_i):

    new_images = 0
    print("per class data sampling begins here \n")
    per_class_labels_list_np_array = np.asarray(per_class_labels_list)
    if(incremental_i == 0):
        edge_cl.eval()
        new_images, new_labels = GetusefulData(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl)
        edge_cl.train()
    else:
        for kk in range(0,len(per_class_labels_list)):
            per_class_labels_list[kk] -= (number_of_classes_to_be_added_ever_increment*incremental_i)
        per_class_labels_list_np_array = np.asarray(per_class_labels_list)
        edge_temp_cl.eval()
        new_images, new_labels = GetusefulData(per_class_features, per_class_labels_list_np_array, Batch_size, edge_temp_cl)
        edge_temp_cl.train()
        for kk in range(0,len(new_labels)):
            new_labels[kk] += (number_of_classes_to_be_added_ever_increment*incremental_i)
    
    return new_images, new_labels

#==========MAIN PROGRAM==========
#LOAD TRAIN TEST SPLIT FILE HERE TO KNOW WHICH IMAGE IS IN TRAINING SET AND TESTING SET
train_indexes = [] #this will store indexes that tells at which index image will be in train set
test_indexes = []
#file_path = r"C:\Users\Swaraj\AppData\Local\Programs\Python\Python36\Journal_paper_codes\SqueezeNet\CUB_200_2011\train_test_split.txt"
file_path = '/home/swarajdubey611/CUB_200_2011/train_test_split.txt'
f = open(file_path, 'r')
lines = f.readlines()
f.close()
train_counter = 0
test_counter = 0
for i in range(len(lines)):
    ind = int(lines[i].split(" ")[1][:-1])
    if(ind == 1):
        train_indexes.append(train_counter)
    else:
        test_indexes.append(test_counter)
    train_counter += 1
    test_counter += 1
#print("length of train_indexes: "+str(len(train_indexes)))
#print("length of test_indexes: "+str(len(test_indexes)))
#print(train_indexes)


squeezenet = models.squeezenet1_0(pretrained=True)
squeezenet = squeezenet.to(device)
squeezenet.eval()

#squeezenet = models.shufflenet_v2_x1_0(pretrained=True) #Shufflenet models
#squeezenet = torch.nn.Sequential(*(list(squeezenet.children())[:-1])) #remove fc layer from ShuffleNet
#squeezenet = squeezenet.to(device)
#squeezenet.eval()

edge_cl = EdgeClassifier()
edge_cl = edge_cl.to(device) #MAP THE MODEL ONTO THE GPU
edge_cl.eval()

edge_temp_cl = EdgeTempClassifier()
edge_temp_cl = edge_temp_cl.to(device) #MAP THE MODEL ONTO THE GPU
edge_temp_cl.eval()

edge_criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
edge_optimizer = optim.Adam(edge_cl.parameters(), lr=1e-4) #ADAM OPTIMIZER


training_image_list = 0
training_image_label_list = []
testing_image_list = 0
testing_image_label_list = []
#must use the absolute path
directory = '/home/swarajdubey611/CUB_200_2011/Images'#r"C:\Users\Swaraj\AppData\Local\Programs\Python\Python36\Journal_paper_codes\SqueezeNet\CUB_200_2011\images"
#directory = '/content/gdrive/My Drive/CUB_200_resized'
index_counter = 0
train_counter = 0
test_counter = 0
class_counter = -1
Exemplar_features = 0
Exemplar_labels = 0
number_of_classes_to_be_added_ever_increment = starting_softmax_neurons
classes_per_training_round_count = 0
edge_incremental_i = 0
number_of_classes_to_be_added_ever_increment = starting_softmax_neurons
total_nerons_in_final_layer= starting_softmax_neurons

#Variables for SMOTE
useful_images_list = []
useful_labels_list = []
length_of_dataset_after_filtration = []

Batch_size = 32
criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
learning_rate = 1e-4
optimizer = optim.Adam(edge_cl.parameters(), lr=learning_rate) #ADAM OPTIMIZER
total_epochs = 15
batch_size = 128 #this batch size is only for data loading

training_features_list = 0
testing_features_list = 0
training_features_list_labels = []
testing_features_list_labels = []
folder_count = 0
train_batch_counter = 0
test_batch_counter = 0
all_testing_features_so_far = 0
all_testing_labels_so_far = []
edge_processing_time_list = []

samples_transmitted_count = 0
length_so_far = 0
number_of_overloads = 0
datasize_sent_list = []
overall_string_tensor = ""
testing_overall_string_tensor = ""
total_append_operation_time = 0

new_images_length_counter = 0
new_images_length_list = []

total_classes = 200
number_of_increments = int(total_classes/starting_softmax_neurons) #total number of classes to be learnt
#number_of_increments = 2
img_dimension = 90
overall_tic = time.time()
for subdir in os.listdir(directory):
    #print("edge_incremental_i: "+str(edge_incremental_i)+", number_of_increments: "+str(number_of_increments))
    if(edge_incremental_i < total_classes):# (classes_per_training_round_count < number_of_classes_to_be_added_ever_increment) and (incremental_i < number_of_increments)
        print(subdir)
        subdir_absolute_path = directory + '/' + subdir
        class_counter += 1
        folder_count += 1
        
        for f in os.listdir(subdir_absolute_path):
            read_tic = time.time()
            ori_img = Image.open(os.path.join(subdir_absolute_path,f)) #this is the JPEG image
            newsize = (img_dimension, img_dimension) 
            ori_img = ori_img.resize(newsize) 
            img = transform(ori_img)
            if(img.shape[0] == 3):#accept only RGB images
                img = img.reshape((1,3,img_dimension,img_dimension)) #reshape the image tensor
                if(index_counter in train_indexes): #training images
                    if(train_counter == 0): #first time appending
                        training_image_list = img
                        train_counter = 1
                    else:
                        training_image_list = np.concatenate((training_image_list,img),axis=0)
                    training_image_label_list.append(class_counter)
                    
                else: #testing images
                    if(test_counter == 0): #first time appending
                        testing_image_list = img
                        test_counter = 1
                    else:
                        testing_image_list = np.concatenate((testing_image_list,img),axis=0)
                    testing_image_label_list.append(class_counter)
                index_counter += 1
            read_toc = time.time()
            
            
        if(folder_count == 1): #Send append message to the cloud
            
            print("Training images size: "+str(len(training_image_label_list)))
            print("Testing images size: "+str(len(testing_image_label_list)))
            print("Starting data sampling")
            folder_count = 0
            edge_tic = time.time()
            #OBTAIN FEATURES OF ALL SAMPLES IN THE SUB FOLDER
            trail_train_dataset = TensorDataset( Tensor(training_image_list), Tensor(np.asarray(training_image_label_list)) )
            Trialtrainloader = DataLoader(trail_train_dataset, batch_size= batch_size, shuffle=False)
            training_image_list = 0
            training_image_label_list = []

            per_class_features = 0
            per_class_features_counter = 0
            per_class_labels_list = []
            for i, data in enumerate(Trialtrainloader, 0):
                inputs, _ = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),3,img_dimension,img_dimension) 
                inputs=inputs.to(device)

                for _ in range(len(inputs)):
                    per_class_labels_list.append(class_counter)

                with torch.no_grad():
                    outputs = squeezenet.features(inputs)
                    print("outputs size: "+str(outputs.shape))
                    if(per_class_features_counter==0):
                        per_class_features = outputs
                        per_class_features_counter = 1
                    else:
                        per_class_features = torch.cat([per_class_features,outputs],dim=0)


            #Carry out data sampling here
            per_class_labels_list_np_array = np.asarray(per_class_labels_list)
            new_images, new_labels = DataSampling(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl, edge_temp_cl,number_of_classes_to_be_added_ever_increment,edge_incremental_i)
            
            new_images_length_counter += len(new_images)
            print("per class data sampling ends here \n")
            Trainloader_iterations = 0
            if(len(new_images) % batch_size == 0):
                Trainloader_iterations = len(new_images) // batch_size
            else:
                Trainloader_iterations = (len(new_images) // batch_size) + 1
            edge_toc = time.time()
            edge_processing_time_list.append(edge_toc-edge_tic)

            samples_transmitted_count += len(new_images)
            
            #Convert the filtered data samples to string format
            traindataset_new = TensorDataset( Tensor(new_images), Tensor(new_labels))
            Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)

            ti = 0
            for i, data in enumerate(Trainloader, 0): #This will only run once i.e. only 1 epoch        
                mini_batch_tic = time.time()
                inputs, labels = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                outputs = inputs

                #Send the data to the cloud device (images and the labels)
                string_tensor = TensorToString(labels,outputs)
                overall_string_tensor = overall_string_tensor + string_tensor + "k"
                mini_batch_toc = time.time()
                time_diff = mini_batch_toc - mini_batch_tic

                if(i==0):
                    ti = time_diff
                else:
                    if((time_diff - ti) > 0.5):
                        print("RAM overload!")
                        if(i == Trainloader_iterations - 1): #RAM overload at the last iteration! so start training already
                            overall_string_tensor = overall_string_tensor[:-1] + "t!"
                            append_ticc = time.time()
                            Cloud(overall_string_tensor)
                            append_tocc = time.time()
                            total_append_operation_time += (append_tocc-append_ticc)
                        else:
                            overall_string_tensor = overall_string_tensor[:-1] + "a!" #remove the last k and add a! at the end
                            append_ticc = time.time()
                            Cloud(overall_string_tensor)
                            append_tocc = time.time()
                            total_append_operation_time += (append_tocc-append_ticc)
                        length_so_far += len(overall_string_tensor)
                        length_so_far *= 1e-6
                        datasize_sent_list.append(length_so_far)
                        overall_string_tensor = ""
                        number_of_overloads += 1


            #get testing features of images and only save that
            ti = 0
            trail_test_dataset = TensorDataset( Tensor(testing_image_list), Tensor(np.asarray(testing_image_label_list)) )
            Trialtestloader = DataLoader(trail_test_dataset, batch_size= batch_size, shuffle=False)
            testing_image_list = 0
            testing_image_label_list = []
            for i, data in enumerate(Trialtestloader, 0):
                mini_batch_tic = time.time()

                inputs, labels = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),3,img_dimension,img_dimension)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                with torch.no_grad():
                    outputs = squeezenet.features(inputs)
                string_tensor = TensorToString(labels,outputs)
                testing_overall_string_tensor = testing_overall_string_tensor + string_tensor + "k"
                mini_batch_toc = time.time()
                time_diff = mini_batch_toc - mini_batch_tic
                if(i==0):
                    ti = time_diff
                else:
                    if((time_diff - ti) > 0.5):
                        print("RAM overload!")
                        if(classes_per_training_round_count == number_of_classes_to_be_added_ever_increment-1):
                            testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "c!"
                            Cloud(testing_overall_string_tensor)
                        else:
                            testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "h!"
                            Cloud(testing_overall_string_tensor)
                        testing_overall_string_tensor = ""

            #reset all the variables
            training_image_list = 0
            testing_image_list = 0
            train_counter = 0
            test_counter = 0
            training_image_label_list = []
            testing_image_label_list = []


        if(classes_per_training_round_count == number_of_classes_to_be_added_ever_increment-1): #Send to the cloud for training

            new_images_length_list.append(samples_transmitted_count)
            samples_transmitted_count = 0
            print("before training edge_incremental_i: "+str(edge_incremental_i))
            if(len(overall_string_tensor)>1):
                overall_string_tensor = overall_string_tensor[:-1] + "t!" #remove the last k and add t! at the end
                length_so_far += len(overall_string_tensor)
                length_so_far *= 1e-6
                datasize_sent_list.append(length_so_far)
                Cloud(overall_string_tensor)        
            overall_string_tensor = ""            


            #Start the classification process
            if(len(testing_overall_string_tensor)>1):
                testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "c!"
                Cloud(testing_overall_string_tensor)
            testing_overall_string_tensor = ""

            #Modify the edge_cl
            print("modify softmax layer edge "+str(edge_cl))
            #Add neurons to the SoftMax classification layer
            
            current_weights = edge_cl.softmax_layer.weight.data
            current_bias = edge_cl.softmax_layer.bias.data

            new_total_nerons_in_final_layer = total_nerons_in_final_layer + number_of_classes_to_be_added_ever_increment
            total_nerons_in_final_layer += number_of_classes_to_be_added_ever_increment
            edge_cl.softmax_layer = nn.Linear(feature_extractor_size,total_nerons_in_final_layer)

            new_weights = -0.0001 * np.random.random_sample((number_of_classes_to_be_added_ever_increment, feature_extractor_size)) + 0.0001 #RANDOMLY INITIALIZED WEIGHTS FOR THE NEW TASKS
            new_weights = torch.from_numpy(new_weights).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
            new_weights = new_weights.to(device)
            new_weights = torch.cat([current_weights,new_weights],dim=0) #Concatenate the new initialized weights with the old weights for old task
            edge_cl.softmax_layer.weight.data.copy_(torch.tensor(new_weights, requires_grad=True))

            new_biases = np.zeros(number_of_classes_to_be_added_ever_increment)
            new_biases = torch.from_numpy(new_biases).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
            new_biases = new_biases.to(device)
            new_biases = torch.cat([current_bias,new_biases],dim=0)
            edge_cl.softmax_layer.bias.data.copy_(torch.tensor(new_biases, requires_grad=True))

            #After adding neurons to the final layer, the network needs to be pushed to the GPU again
            edge_cl = edge_cl.to(device)


            #Assign random weights to tempcl
            edge_modifiable_cl = copy.deepcopy(edge_cl)
            edge_modifiable_cl_weight = edge_modifiable_cl.softmax_layer.weight[(total_nerons_in_final_layer - number_of_classes_to_be_added_ever_increment):total_nerons_in_final_layer] #base weights
            edge_modifiable_cl_bias = edge_modifiable_cl.softmax_layer.bias[(total_nerons_in_final_layer - number_of_classes_to_be_added_ever_increment):total_nerons_in_final_layer] #base biases
            edge_temp_cl.softmax_layer.weight.data.copy_(torch.tensor(edge_modifiable_cl_weight, requires_grad=True))
            edge_temp_cl.softmax_layer.bias.data.copy_(torch.tensor(edge_modifiable_cl_bias, requires_grad=True))

            #reset all the variables for the next incremental training round
            training_features_list = 0
            training_features_list_labels = []
            testing_features_list = 0
            testing_features_list_labels = []
            train_batch_counter = 0
            test_batch_counter = 0
            edge_incremental_i += 1
            classes_per_training_round_count = -1

            print("after training edge_incremental_i: "+str(edge_incremental_i))
        else: #Send to the cloud as append append
            print("acloud append")
            if(len(overall_string_tensor)>1):
                overall_string_tensor = overall_string_tensor[:-1] + "a!" #remove the last k and add a! at the end
                length_so_far += len(overall_string_tensor)
                length_so_far *= 1e-6
                datasize_sent_list.append(length_so_far)
                append_ticc = time.time()
                Cloud(overall_string_tensor)
                append_tocc = time.time()
                total_append_operation_time += (append_tocc-append_ticc)
            overall_string_tensor = ""

            if(len(testing_overall_string_tensor)>1):
                testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "h!"
                Cloud(testing_overall_string_tensor)
            testing_overall_string_tensor = ""

    classes_per_training_round_count += 1

stop_cloud_message = "d!"
Cloud(stop_cloud_message)
overall_toc = time.time()
print("=========================EDGE RESULTS===========================")
print("overall time: "+str(overall_toc - overall_tic))
print("new_images_length_list: "+str(new_images_length_list))
print("total images sent: "+str(sum(new_images_length_list)))
print("amount of data sent so far (MB): ")
print(datasize_sent_list)
print("TOTAL data sent (MB): ")
print(sum(datasize_sent_list))
