import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import time as time
from torch.utils.data import DataLoader, TensorDataset
import numpy as np
from torch import Tensor
import random 
from torchvision import datasets, models
import statistics
import collections
import copy
import math
import requests
import socket
import os
from PIL import Image
import scipy.stats as ss
from scipy.stats import norm
from scipy.stats import median_test
from scipy.io import loadmat
from scipy.stats import rankdata
from scipy import stats

#from google.colab import drive
#drive.mount('/content/gdrive')

transform = transforms.Compose(
    [transforms.ToTensor()
    ])

#========DEFINE THE CNN MODEL=====
starting_softmax_neurons = 40
feature_d = 1280
feature_w = 3
feature_h = 3

feature_extractor_size = feature_d*feature_w*feature_h
#second_last_layer_size = 1000
class CloudClassifier(nn.Module):
    def __init__(self):
        super(CloudClassifier, self).__init__()
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons) #2 output nodes initially
        self.drop1 = nn.Dropout(0.30)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop1(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
print("device: "+str(device))
cloud_cl = CloudClassifier()
cloud_cl = cloud_cl.to(device)

#FUNCTION TO CONVERT STRING BACK TO TENSOR (LABELS AND THE FEATURE MAPS)
def DecomposeString(Feature,batch_size):
    string_tensor = Feature
    listt = string_tensor.split(",")
    arr = np.array(listt,dtype = float)

    depth = int(arr[0])
    width = int(arr[1])
    height = int(arr[2])

    batch_size = int(arr.shape[0]/(depth*width*height))
    
    labels = []
    for i in range(3,3+batch_size):
        labels.append(arr[i])

    torch_labels = torch.from_numpy(np.array(labels,dtype=int))
    #remove first 3 elements from list
    arr = arr[3+batch_size:]

    torch_tensor = torch.from_numpy(arr)
    torch_tensor = torch_tensor.view(-1,depth,width,height)
    return torch_tensor,torch_labels

#Function to create artificial samples
def SMOTE(img_train, lab_train, total_new_points):
    length_of_dataset = len(img_train)
    number_of_points_to_be_chosen = 2 #distance can be calculated using only 2 points

    number_of_data_created = 0
    new_data_img = 0
    new_data_lab = []
    while(number_of_data_created != total_new_points):
        sequence = [i for i in range(length_of_dataset)]
        point_indices = random.sample(sequence, number_of_points_to_be_chosen) #choose any 2 training images from img_train
        
        p1 = img_train[point_indices[0]]
        p2 = img_train[point_indices[1]]

        new_lab = lab_train[point_indices[0]] #choose any point as it will have same index

        number_of_points = 5 if (total_new_points - number_of_data_created) > 5 else (total_new_points - number_of_data_created)
        for i in range(0,number_of_points):
            lamda = np.random.uniform(0,1)
            new_point = p1 + (lamda*(p2-p1))
            new_point = np.reshape(new_point,(1,feature_w,feature_h,feature_d))
            if(number_of_data_created == 0):
                new_data_img = new_point
                new_data_lab.append(new_lab)
            else:
                new_data_img = np.concatenate((new_data_img, new_point), axis=0)
                new_data_lab.append(new_lab)
            #print("new_data_img_shape: "+str(new_data_img.shape))
            number_of_data_created += 1

    new_data_lab = np.asarray(new_data_lab)
    #new_data_img = np.concatenate((new_data_img, img_train), axis=0)
    #new_data_lab = np.concatenate((new_data_lab, lab_train), axis=0)

    return new_data_img,new_data_lab

    """
    #RANDOM OVERSAMPLING (SIMPLY DUPLICATING SOME OF THE FEATURES)
    randomlist = random.sample(range(0, length_of_dataset), total_new_points)
    random_features = img_train[randomlist]
    random_features_labels = lab_train[randomlist]

    #APPEND THE NEW DATA AND THE RESPECTIVE SAMPLES
    new_features = np.concatenate((img_train, random_features), axis=0)
    new_features_labels = lab_train + random_features_labels
    return new_features, new_features_labels
    #print(randomlist)
    """

#FUNCTION TO CHECK IF SPARSITY REALLY WORKS OR NOT
def CheckSparsity(cloud_cl):
    useful_weights_count = 0
    total_weights = 0
    for name,layer in cloud_cl.named_modules():
        if(isinstance(layer,nn.Linear)):
            weight_tensor = layer.weight.data #Get weight matrix of the nn.Linear layer
            x = torch.zeros(weight_tensor.shape)
            x[weight_tensor != 0] = 1 #all elements not equal to 0 are assigned as 1 in tensor `x`
            #sparsity_count += (torch.numel(x) - torch.count_nonzero(x)).item()
            useful_weights_count += torch.count_nonzero(x).item()
            total_weights += torch.numel(x)
    print("useful_weights_count: "+str(useful_weights_count))
    pruned_weights_count = total_weights - useful_weights_count
    pruned_weights_percent = (pruned_weights_count/total_weights)*100.0
    return total_weights, pruned_weights_count, pruned_weights_percent

def GetTrainingAccuracy(cl,device,Trainloader):
    cl.eval()
    correct = 0
    total = 0
    top = 5
    with torch.no_grad():
        for data in Trainloader:
            images, labels = data

            images = images.clone().detach()
            images = images.view(len(images),feature_d,feature_w,feature_h) #RESHAPE THE IMAGES
            labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

            images=images.to(device)
            labels=labels.to(device)

            outputs = cl.forward(images)

            #COMPUTE TOP-5 accuracy
            _,pred=outputs.topk(top,1,largest=True,sorted=True) #select top 5
            total += labels.size(0)

            for j in range(len(pred)):
                tensor = pred[j]
                true_label = labels[j]
                for k in range(len(tensor)):
                    if(true_label == tensor[k].item()):
                        correct += 1
                        break

    acc = float(correct/total)*100.0
    cl.train()
    return acc


def RandomPruning(cloud_cl,device,Trainloader):
    
    acceptable_training_loss = 0.5
    print("Random pruning prcoess starting")
    total_nn_weights, pruned_weights, pruned_weights_percent = CheckSparsity(cloud_cl)
    print("pruned_weights_percent: "+str(pruned_weights_percent))

    #Get baseline accuracy
    baseline_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)

    best_models = []
    #Prune for every layer
    for name,layer in cloud_cl.named_modules():
        if(isinstance(layer,nn.Linear)):
            weight_tensor = copy.deepcopy(layer.weight.data) #Get weight matrix of the nn.Linear layer
            tensor_size = 1
            for i in range(len(weight_tensor.shape)):
                tensor_size *= weight_tensor.shape[i]
            flattened_vector = weight_tensor.view(-1,tensor_size)

            sparsity_increment = 1 #1%
            sparsity_percent = 0.0 #0%
            weights_before_sparsification = 0
            weights_before_sparsification = copy.deepcopy(layer.weight.data) #save the weights of layer l before going sparse
            for i in range(0,100,sparsity_increment):
                sparsity_percent += i #This is the new sparsity level
                sparsity_percent /= 100.0 #convert in range 0 to 1
                print("sparsity_percent is: "+str(sparsity_percent))
                no_of_elements_to_be_sparse = int(sparsity_percent * tensor_size) #number of weights to be pruned
                if(no_of_elements_to_be_sparse > 0): #Sparsity only takes place if the
                    random_indices = random.sample(range(0, tensor_size-1), no_of_elements_to_be_sparse)
                    random_indices = torch.LongTensor(random_indices) #indices to be made sparse
                    
                    weight_tensor = weight_tensor.view(-1,tensor_size) #flatten the weight matrix
                    weight_tensor[0][random_indices] = 0 #prune the parameters
                    layer.weight.data = weight_tensor.view(layer.weight.data.shape)
                best_models.append(cloud_cl)
                temp_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)
            
                print("temp_accuracy: "+str(temp_accuracy))
                if(abs(baseline_accuracy - temp_accuracy) > acceptable_training_loss):
                    if(len(best_models)>=2): #just to ensure no crashing takes place
                        cloud_cl = best_models[-2]
                        print("accuracy loss exceeded, breaking now and check for next layer")
                        break
                else: #model did not prune well enough so restart again
                    layer.weight.data = weights_before_sparsification
                    print("refreshed the layer l weights")

    total_nn_weights, pruned_weights, pruned_weights_percent = CheckSparsity(cloud_cl)
    print("pruned_weights_percent: "+str(pruned_weights_percent))

    return cloud_cl, total_nn_weights, pruned_weights, pruned_weights_percent


def MagnitudePruning(cloud_cl,device,Trainloader):
    acceptable_training_loss = 0.5

    print("magnitude pruning prcoess starting")
    total_nn_weights, pruned_weights, pruned_weights_percent = CheckSparsity(cloud_cl)
    print("pruned_weights_percent: "+str(pruned_weights_percent))

    #Get baseline accuracy
    baseline_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)

    best_models = []
    #Prune for every layer
    for name,layer in cloud_cl.named_modules():
        if(isinstance(layer,nn.Linear)):
            weight_tensor = layer.weight.data #Get weight matrix of the nn.Linear layer
            tensor_size = 1
            for i in range(len(weight_tensor.shape)):
                tensor_size *= weight_tensor.shape[i]
            flattened_vector = weight_tensor.view(-1,tensor_size)
            sorted_tensors, _ = torch.sort(flattened_vector,descending=False)

            sparsity_increment = 1 #1%
            sparsity_percent = 0.0 #0%
            for i in range(0,100,sparsity_increment):
                sparsity_percent += i #This is the new sparsity level
                sparsity_percent /= 100.0 #convert in range 0 to 1
                print("sparsity_percent is: "+str(sparsity_percent))
                no_of_elements_to_be_sparse = int(sparsity_percent * tensor_size) #number of weights to be pruned
                if(no_of_elements_to_be_sparse > 0): #Sparsity only takes place if the
                    cut_off_weight = sorted_tensors[0][no_of_elements_to_be_sparse-1].item()
                    layer.weight.data = torch.where(weight_tensor.double() > cut_off_weight, weight_tensor.double(), 0.0).float()
                best_models.append(cloud_cl)
                temp_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)
                print("temp_accuracy: "+str(temp_accuracy))
                if(abs(baseline_accuracy - temp_accuracy) > acceptable_training_loss):
                    if(len(best_models)>=2): #just to ensure no crashing takes place
                        cloud_cl = best_models[-2]
                        print("accuracy loss exceeded, breaking now")
                        break

    total_nn_weights, pruned_weights, pruned_weights_percent = CheckSparsity(cloud_cl)
    print("pruned_weights_percent: "+str(pruned_weights_percent))

    return cloud_cl, total_nn_weights, pruned_weights, pruned_weights_percent


def GetThresholds(weight_difference_tensor):
    thresholds = []
    quantiles = 20
    tensor_size = 1
    for i in range(len(weight_difference_tensor.shape)):
        tensor_size *= weight_difference_tensor.shape[i]
    flattened_vector = weight_difference_tensor.view(-1,tensor_size)
    sorted_tensors, _ = torch.sort(flattened_vector,descending=False)

    print("sorted_tensors: "+str(sorted_tensors))
    index_of_first_positive_element = (sorted_tensors < 0).sum().item()
    #index_of_first_positive_element = 0 #This is just to replicate the original fitcnn algorithm

    spacing = int(len(sorted_tensors[0])/quantiles)
    for i in range(index_of_first_positive_element,len(sorted_tensors[0]),spacing):
        thresholds.append(sorted_tensors[0][i].item())
    
    print("length of thresholds: "+str(len(thresholds)))
    return thresholds


def SpecialThresholds(thresholds):
    special_threshold_indices = []
    quarter_index_threshold = thresholds[int(0.25*len(thresholds))]
    half_index_threshold = thresholds[int(0.50*len(thresholds))]
    three_quarter_index_threshold = thresholds[int(0.75*len(thresholds))]
    special_threshold_indices.append(quarter_index_threshold)
    special_threshold_indices.append(half_index_threshold)
    special_threshold_indices.append(three_quarter_index_threshold)
    return special_threshold_indices

def ImprovedJuicer(cloud_cl,pre_training_cl,post_training_cl,Trainloader,device):

    acceptable_training_accuracy_loss = 0.5
    baseline_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)
    print("baseline_accuracy: "+str(baseline_accuracy))
    total_useful_weights = 0
    total_nn_weights = 0
    temp_model = copy.deepcopy(post_training_cl)
    ann_layer_counter = -1
    for (module1,module2) in zip(pre_training_cl.modules(),post_training_cl.modules()):
        if((isinstance(module1,nn.Linear)) and (isinstance(module2,nn.Linear))):
            print("EXTRACTING WEIGHTS FOR LAYER: "+str(ann_layer_counter))
            post_weight, pre_weight = module2.weight.data, module1.weight.data
            weight_tensor_difference = abs(post_weight - pre_weight) #get layer l weight difference
            thresholds = GetThresholds(weight_tensor_difference) #get all thresholds for layer l
            #print("before length of thresholds: "+str(len(thresholds)))

            
            layer_l_best_weights_list = []
            special_threshold_indices = SpecialThresholds(thresholds) #get thresholds ar 25%, 50% and 75% positions
            starting_threshold_index = 0
            for i in range(len(special_threshold_indices)):
                threshold = special_threshold_indices[i]
                #print("shortcut checking for special threshold: "+str(threshold))
                layer_l_weights = torch.where(weight_tensor_difference.double() > threshold, post_weight.double(), pre_weight.double()).float()
                temp_model_layer_counter = -1
                for name,layer in temp_model.named_modules():
                    if((isinstance(layer,nn.Linear)) and (ann_layer_counter == temp_model_layer_counter)):
                        #print("special threshold inside tempmodel")
                        layer.weight.data = layer_l_weights
                        layer_l_best_weights_list.append(layer_l_weights)
                        break
                    temp_model_layer_counter += 1
                temp_accuracy = GetTrainingAccuracy(temp_model,device,Trainloader) #Calculate the training accuracy on temp_model
                if(abs(baseline_accuracy - temp_accuracy)>=acceptable_training_accuracy_loss):
                    #print("no shortcuts available..")
                    break
                else:
                    starting_threshold = threshold
                    starting_threshold_index = thresholds.index(starting_threshold)
                    #print("starting_threshold is: "+str(starting_threshold))
                    #print("starting_threshold_index is: "+str(starting_threshold_index))
            thresholds = thresholds[starting_threshold_index:] #This is the fine tuned threshold list
            print("length of thresholds now: "+str(len(thresholds))+"\n")
            

            total_nn_weights += torch.numel(weight_tensor_difference)
            layer_l_best_weights_list = []
            for threshold in thresholds:
                print("check for threshold: "+str(threshold))
                layer_l_weights = torch.where(weight_tensor_difference.double() > threshold, post_weight.double(), pre_weight.double()).float()
                
                temp_model_layer_counter = -1
                for name,layer in temp_model.named_modules():
                    if((isinstance(layer,nn.Linear)) and (ann_layer_counter == temp_model_layer_counter)):
                        print("inside tempmodel")
                        layer.weight.data = layer_l_weights
                        layer_l_best_weights_list.append(layer_l_weights)
                        break
                    temp_model_layer_counter += 1

                print("getting training accuracy")
                temp_accuracy = GetTrainingAccuracy(temp_model,device,Trainloader) #Calculate the training accuracy on temp_model
                print("temp_accuracy: "+str(temp_accuracy))
                if(abs(baseline_accuracy - temp_accuracy)>=acceptable_training_accuracy_loss):
                    #get the previous temp_model
                    layer_l_best_weights = layer_l_best_weights_list[-2]
                    total_useful_weights += (module2.weight.data == layer_l_best_weights).sum().item()
                    module2.weight.data = layer_l_best_weights

                    #Assign layer_l_best_weights to layer l of temp_model
                    temp_model_layer_counter = -1
                    for name,layer in temp_model.named_modules():
                        if((isinstance(layer,nn.Linear)) and (ann_layer_counter == temp_model_layer_counter)):
                            print("Again temp_model_layer_counter: "+str(temp_model_layer_counter))
                            layer.weight.data = layer_l_best_weights
                            break
                        temp_model_layer_counter += 1

                    layer_l_best_weights_list = []
                    break
        ann_layer_counter += 1

    #useful_weights_count, total_weights, pruned_weights_percent = CheckSparsity(post_training_cl)
    pruned_weights = total_nn_weights - total_useful_weights
    pruned_weights_percent = (pruned_weights/total_nn_weights)*100.0
    print("total_useful_weights: "+str(total_useful_weights))
    print("pruned_weights_percent: "+str(pruned_weights_percent))
    return post_training_cl, total_nn_weights, pruned_weights, pruned_weights_percent


test_features_first_time = True
appending_message = False
training_features = 0
training_features_labels = 0
training_features_so_far = 0
training_features_labels_so_far = 0

testing_features = 0
testing_features_labels = 0
testing_features_so_far = 0
testing_features_labels_so_far = 0

#Variables for SMOTE
useful_images_list = []
useful_labels_list = []
length_of_dataset_after_filtration = []
acc_list = []

ori_weight_message_string_length_list = []

training_time_list = []
weight_message_string_length = []
classification_time_list = []
append_time_list = []

incremental_training_accuracy_list = []
smote_time_list = []

pruning_time_list = []
pruned_weights_percent_list = []
pruned_weights_list = []
total_nn_weights_list = []

Exemplar_features = 0
Exemplar_labels = 0
first_time = 0
incremental_i = 0
cloud_total_nerons_in_final_layer = starting_softmax_neurons
#cloud_temp_cl = CloudTempClassifier()

def Cloud(Feature):
    global test_features_first_time
    global appending_message
    global training_features
    global training_features_labels
    global training_features_so_far
    global training_features_labels_so_far
    global testing_features
    global testing_features_labels
    global testing_features_so_far
    global testing_features_labels_so_far
    global ori_weight_message_string_length_list

    global training_time_list
    global weight_message_string_length
    global classification_time_list
    global append_time_list

    #Variables for SMOTE
    global useful_images_list
    global useful_labels_list
    global length_of_dataset_after_filtration
    global smote_time_list

    global Exemplar_features
    global Exemplar_labels
    global first_time
    global incremental_i
    global acc_list
    global cloud_total_nerons_in_final_layer
    global cloud_cl
    global incremental_training_accuracy_list
    
    global pruning_time_list
    global pruned_weights_percent_list
    global pruned_weights_list
    global total_nn_weights_list

    batch_size = 32
    cloud_criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
    cloud_special_criterion = nn.CrossEntropyLoss(reduction='none')
    cloud_learning_rate = 0.0002
    cloud_training_epochs = 30
    cloud_optimizer = optim.Adam(cloud_cl.parameters(), lr=cloud_learning_rate) #ADAM OPTIMIZER

    Feature = Feature.replace(" ","") #remove any spaces
    task_character = Feature[-2] #Check the 2nd last character and decide what the program should do now
    print("task_character: "+str(task_character)+"\n")

    if(task_character == 't'): #train the model

        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k') #every feature map is separated by 'k'

        Trainloader = 0
        print("start training \n")
        #All features received from the IoT edge device
        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)

            if(i==0):
                training_features = inputs
                training_features_labels = labels
            else:
                training_features = torch.cat([training_features,inputs],dim=0)
                training_features_labels = torch.cat([training_features_labels,labels],dim=0)
        
        training_features = training_features.float()
        training_features_labels = training_features_labels.float()

        #Joing all the tensors that were received due to RAM overload on IoT edge device
        if(appending_message == True): #This means IoT device has sent a message due to RAM overload
            training_features = torch.cat([training_features,training_features_so_far],dim=0)
            training_features_labels = torch.cat([training_features_labels,training_features_labels_so_far],dim=0)
            appending_message = False
        
        print("training_features before training: "+str(training_features.shape))
        training_features = np.reshape(training_features,(len(training_features),feature_w,feature_h,feature_d))

        #Exemplar loader
        exemplardataset = TensorDataset( Tensor(training_features), Tensor(training_features_labels) )
        Exemplarloader = DataLoader(exemplardataset, batch_size= batch_size, shuffle=False)

        if(first_time==0):

            length_of_dataset_after_filtration.append(len(training_features)) #LENGTH OF SAMPLES IN THE RESPECTIVE TRAINING ROUND
            useful_images_list.append(training_features)
            useful_labels_list.append(training_features_labels)
            
            traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
            Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
        else:
            #CARRY OUT SMOTE OVER HERE
            if(len(training_features)<=length_of_dataset_after_filtration[incremental_i-1]):
                print("\ncurrent incremental training round in minority ")
                if(len(training_features)<length_of_dataset_after_filtration[incremental_i-1]):
                    total_new_points = length_of_dataset_after_filtration[incremental_i-1] - len(training_features)
                    smote_tic = time.time()
                    remaining_images,remaining_labels = SMOTE(training_features, training_features_labels, total_new_points)
                    smote_toc = time.time()
                    smote_time = smote_toc - smote_tic
                    smote_time_list.append(smote_time)
                    #remaining_images = np.reshape(remaining_images,(total_new_points,feature_extractor_size,w,h))
                    remaining_images = np.reshape(remaining_images,(total_new_points,feature_w,feature_h,feature_d))

                    print("training_features shape: "+str(training_features.shape))
                    print("remaining_images shape: "+str(remaining_images.shape))
                    training_features = np.concatenate((training_features,remaining_images),axis=0)
                    training_features_labels = np.concatenate((training_features_labels,remaining_labels),axis=0)
                    traindataset = TensorDataset( Tensor(training_features), Tensor(training_features_labels) )
                    Trainloader_trail = DataLoader(traindataset, batch_size= batch_size, shuffle=False)
                    #get initial losses
                    traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
                    Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
                    #length_of_dataset_after_filtration[incremental_i-1] = len(training_features)

                    print("length of new images: "+str(len(training_features)))
                    length_of_dataset_after_filtration.append(len(training_features))
                    useful_images_list.append(training_features)
                    useful_labels_list.append(training_features_labels)
                else:
                    print("same number of images as last time")
                    length_of_dataset_after_filtration.append(len(training_features))
                    useful_images_list.append(training_features)
                    useful_labels_list.append(training_features_labels)

                    traindataset_new = TensorDataset( Tensor(training_features), Tensor(training_features_labels))
                    Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
            elif(len(training_features)>length_of_dataset_after_filtration[incremental_i-1]):
                n_images_total = 0
                n_labels_total = 0
                print("length of useful_images_list: "+str(len(useful_images_list)))
                for kk in range(0,len(useful_images_list)):
                    total_new_points = len(training_features) - len(useful_images_list[kk])
                    smote_tic = time.time()
                    n_images,n_labels = SMOTE(useful_images_list[kk], useful_labels_list[kk], total_new_points)
                    smote_toc = time.time()
                    smote_time = smote_toc - smote_tic
                    smote_time_list.append(smote_time)
                    n_images = np.reshape(n_images,(total_new_points,feature_w,feature_h,feature_d))
                    print("useful_images_list[kk] shape: "+str(useful_images_list[kk].shape))
                    print("n_images(new artifical images) shape: "+str(n_images.shape))
                    useful_images_list[kk] = np.concatenate((useful_images_list[kk],n_images),axis=0)
                    useful_labels_list[kk] = np.concatenate((useful_labels_list[kk],n_labels),axis=0)
                    length_of_dataset_after_filtration[kk] = len(useful_images_list[kk])

                    if(kk==0):
                        n_images_total = n_images
                        n_labels_total = n_labels
                    else:
                        n_images_total = np.concatenate((n_images_total,n_images),axis=0)
                        n_labels_total = np.concatenate((n_labels_total,n_labels),axis=0)
                
                print("n_images_total(TOTAL artificial samples) shape: "+str(n_images_total.shape))
                print("current images_train shape: "+str(training_features.shape))
                n_images_total = np.concatenate((n_images_total,training_features),axis=0)
                n_labels_total = np.concatenate((n_labels_total,training_features_labels),axis=0)
                print("n_images_total(TOTAL appended samples) shape: "+str(n_images_total.shape))

                traindataset_new = TensorDataset( Tensor(n_images_total), Tensor(n_labels_total) )
                Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)
                

                length_of_dataset_after_filtration.append(len(training_features))
                useful_images_list.append(training_features)
                useful_labels_list.append(training_features_labels)

            current_features_train = 0
            current_features_label = 0

            #Get features of the current class
            for i, data in enumerate(Trainloader, 0):
                inputs, labels = data
                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                Features = inputs
                #Features = Features.to(device)
                #labels = labels.to(device)

                if(i==0):
                    current_features_train = Features
                    current_features_label = labels
                else:
                    current_features_train = torch.cat([current_features_train,Features],dim=0)
                    current_features_label = torch.cat([current_features_label,labels],dim=0)

            #APPEND THE CURRENT SAMPLES FEATURES WITH THE EXEMPLARS
            appended_dataset_image = torch.cat([Exemplar_features,current_features_train],dim=0)
            appended_dataset_label = torch.cat([Exemplar_labels,current_features_label],dim=0)

            print("appended data shape is: "+str(appended_dataset_image.shape)+" and appended label shape is: "+str(appended_dataset_label.shape)+"\n")
            dataset = TensorDataset( appended_dataset_image, appended_dataset_label )
            Trainloader = DataLoader(dataset, batch_size= batch_size, shuffle=True)

        print("data loader prepared \n")

        pre_training_cl = copy.deepcopy(cloud_cl)
        #pre_training_weight_list = pre_training_cl.softmax_layer.weight.view(-1,feature_extractor_size*cloud_total_nerons_in_final_layer).cpu().detach().numpy().tolist()[0]
        #pre_training_bias_list = pre_training_cl.softmax_layer.bias.view(-1,cloud_total_nerons_in_final_layer).cpu().detach().numpy().tolist()[0]

        print("before training, cloud_cl is: "+str(cloud_cl))
        training_time_tic = time.time()
        #Train all the features on the cloud now
        for epoch in range(0,cloud_training_epochs):
            for i, data in enumerate(Trainloader, 0):

                inputs, labels = data

                #Pass the data now to the neural network
                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                outputs = cloud_cl.forward(inputs)
                loss = cloud_criterion(outputs, labels)
                        
                cloud_optimizer.zero_grad()
                loss.backward()
                cloud_optimizer.step()

        training_time_toc = time.time()
        training_time = (training_time_toc-training_time_tic)
        training_time_list.append(training_time)
        post_training_cl = copy.deepcopy(cloud_cl)

        #incremental_training_accuracy = GetTrainingAccuracy(cloud_cl,device,Trainloader)
        #incremental_training_accuracy_list.append(incremental_training_accuracy)

        print("done training \n")
        training_features = 0
        training_features_labels = 0
        first_time = 1
        
        #juicer algorithm goes here
        pruning_tic = time.time()
        #cloud_cl, total_nn_weights, pruned_weights, pruned_weights_percent = RandomPruning(cloud_cl,device,Trainloader)
        #cloud_cl, total_nn_weights, pruned_weights, pruned_weights_percent = MagnitudePruning(cloud_cl,device,Trainloader) #Magnitude pruning
        #print("before: ")
        #print(cloud_cl.softmax_layer.weight.data)
        cloud_cl, total_nn_weights, pruned_weights, pruned_weights_percent = ImprovedJuicer(cloud_cl,pre_training_cl,post_training_cl,Trainloader,device)
        #print("after: ")
        #print(cloud_cl.softmax_layer.weight.data)
        
        pruned_weights_percent_list.append(pruned_weights_percent)
        pruned_weights_list.append(pruned_weights)
        total_nn_weights_list.append(total_nn_weights)
        pruning_toc = time.time()
        pruning_time = pruning_toc - pruning_tic
        pruning_time_list.append(pruning_time)


        #print("Exemplar loading, incremental_i value: "+str(incremental_i))
        for i, data in enumerate(Exemplarloader, 0):
            inputs, labels = data

            inputs = inputs.clone().detach()
            inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
            labels = labels.long()

            Features = inputs
            #print("Features shape: "+str(Features.shape))

            #print("i = "+str(i)+", incremental_i: "+str(incremental_i))
            #Features = inputs #RAW IMAGES ARE EXEMPLARS 
            if(i==0 and incremental_i==0):
                with torch.no_grad():
                    Exemplar_features = Features
                    Exemplar_labels = labels
                    #print("NN at iter: "+str(i)+" exemplar shape is: "+str(Exemplar_features.shape)+" and label shape is: "+str(Exemplar_labels.shape)+"\n")
            else:
                with torch.no_grad():
                    #print("Exemplar_features shape: "+str(Exemplar_features.shape))
                    Exemplar_features = torch.cat([Exemplar_features,Features],dim=0)
                    Exemplar_labels = torch.cat([Exemplar_labels,labels],dim=0)
                    #print("at iter: "+str(i)+" exemplar shape is: "+str(Exemplar_features.shape)+" and label shape is: "+str(Exemplar_labels.shape)+"\n")

        incremental_i += 1 #INCREMENTAL TRAINING ROUND COUNTER
        #appending_message = False

        #paramter_string_size = len(weight_message_string)*1e-6
        #weight_message_string_length.append(paramter_string_size)

        #print("weight message string: "+str(weight_message_string))

    elif(task_character == 'h'):
        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k')

        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            testing_features = inputs
            testing_features_labels = labels

            testing_features = testing_features.float()
            testing_features_labels = testing_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and test_features_first_time == True): #The very first time we are processing an append message
                test_features_first_time = False
                testing_features_so_far = testing_features
                testing_features_labels_so_far = testing_features_labels
            else:
                testing_features_so_far = torch.cat([testing_features_so_far,testing_features],dim=0)
                testing_features_labels_so_far = torch.cat([testing_features_labels_so_far,testing_features_labels],dim=0)

        print("cloud testing append, testing_features_so_far shape: "+str(testing_features_so_far.shape))

    elif(task_character == 'c'): #time for classification (overall accuracy)
        #optimizer = optim.Adam(cl.parameters(), lr=learning_rate, weight_decay=0.00001) #ADAM OPTIMIZER
        #print("classification start now, the model is: "+str(cloud_cl))
        classification_time_tic = time.time()
        cloud_cl.eval() #put the model is evaluation mode

        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k')


        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            testing_features = inputs
            testing_features_labels = labels

            testing_features = testing_features.float()
            testing_features_labels = testing_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and test_features_first_time == True): #The very first time we are processing an append message
                test_features_first_time = False
                testing_features_so_far = testing_features
                testing_features_labels_so_far = testing_features_labels
            else:
                testing_features_so_far = torch.cat([testing_features_so_far,testing_features],dim=0)
                testing_features_labels_so_far = torch.cat([testing_features_labels_so_far,testing_features_labels],dim=0)

        print("classification testing_features_so_far shape is: "+str(testing_features_so_far.shape))
        testdatasett = TensorDataset( testing_features_so_far, testing_features_labels_so_far )
        Testloaderr = DataLoader(testdatasett, batch_size= batch_size, shuffle=False)

        correct = 0
        total = 0
        top = 5
        print(cloud_cl)
        with torch.no_grad():
            for data in Testloaderr:
                images, labels = data

                images = images.clone().detach()
                images = images.view(len(images),feature_d,feature_w,feature_h) #RESHAPE THE IMAGES
                labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

                images=images.to(device)
                labels=labels.to(device)

                #outputs = squeezenet.features(images)
                outputs = cloud_cl.forward(images)

                #COMPUTE TOP-5 accuracy
                _,pred=outputs.topk(top,1,largest=True,sorted=True) #select top 5
                total += labels.size(0)

                for j in range(len(pred)):
                    tensor = pred[j]
                    true_label = labels[j]
                    for k in range(len(tensor)):
                        if(true_label == tensor[k].item()):
                            correct += 1
                            break
        if(total>0):
            acc = float(correct/total)*100.0
            print("correct is: "+str(correct)+"\n")
            print("total is: "+str(total)+"\n")
            print("accuracy is: "+str(acc)+"\n")
            acc_list.append(acc)

        cloud_cl.train()

        #ADD NEURONS TO THE FINAL SOFTMAX CLASSIFICATION LAYER
        print("modifying the cloud model now \n")
        
        #Add nerons to the fina layer and set the new random weights
        current_weights = cloud_cl.softmax_layer.weight.data
        current_bias = cloud_cl.softmax_layer.bias.data

        cloud_total_nerons_in_final_layer = cloud_total_nerons_in_final_layer + starting_softmax_neurons
        cloud_cl.softmax_layer = nn.Linear(feature_extractor_size,cloud_total_nerons_in_final_layer)

        print("cloud_total_nerons_in_final_layer: "+str(cloud_total_nerons_in_final_layer))

        new_weights = -0.0001 * np.random.random_sample((starting_softmax_neurons, feature_extractor_size)) + 0.0001 #RANDOMLY INITIALIZED WEIGHTS FOR THE NEW TASKS
        new_weights = torch.from_numpy(new_weights).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
        new_weights = new_weights.to(device)
        new_weights = torch.cat([current_weights,new_weights],dim=0) #Concatenate the new initialized weights with the old weights for old task
        cloud_cl.softmax_layer.weight.data.copy_(torch.tensor(new_weights, requires_grad=True))

        new_biases = np.zeros(starting_softmax_neurons)
        new_biases = torch.from_numpy(new_biases).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
        new_biases = new_biases.to(device)
        new_biases = torch.cat([current_bias,new_biases],dim=0)
        cloud_cl.softmax_layer.bias.data.copy_(torch.tensor(new_biases, requires_grad=True))

        #After adding neurons to the final layer, the network needs to be pushed to the GPU again
        cloud_cl = cloud_cl.to(device)
        #print("cloud_cl after appending is: "+str(cloud_cl))

        cloud_optimizer = optim.Adam(cloud_cl.parameters(), lr=cloud_learning_rate) #UPDATE OPTIMIZER

        classification_time_toc = time.time()
        classification_time = (classification_time_toc - classification_time_tic)
        classification_time_list.append(classification_time)

        print("classification done\n")
    
    elif(task_character == 'a'): #just append the tensors and send back acknowledgement back to IoT edge device

        append_tic = time.time()
        #Split 'k' and decompse string ony by one
        Feature = Feature[:-2]
        Feature_list = Feature.split('k') #every feature map is separated by 'k'

        print("start appending \n")
        for i in range(0,len(Feature_list)):
            inputs,labels = DecomposeString(Feature_list[i],batch_size)
            
            training_features = inputs
            training_features_labels = labels

            training_features = training_features.float()
            training_features_labels = training_features_labels.float()

            #print("mid-level appending done")
            if(i == 0 and appending_message == False): #The very first time we are processing an append message
                training_features_so_far = training_features
                training_features_labels_so_far = training_features_labels
            else:
                training_features_so_far = torch.cat([training_features_so_far,training_features],dim=0)
                training_features_labels_so_far = torch.cat([training_features_labels_so_far,training_features_labels],dim=0)

        print("training_features_so_far shape: "+str(training_features_so_far.shape))
        print("done appending")
        appending_message = True
        append_toc = time.time()
        append_time_list.append(append_toc-append_tic)

    elif(task_character == 'd'): #means process done
        print("============================CLOUD RESULTS=====================================")
        print("number of classes being learnt at a time: "+str(starting_softmax_neurons))
        print("overall training_time_list: "+str(training_time_list))
        print("TOTAL TRAINING TIME (SECONDS): "+str(sum(training_time_list)))
        print("Testing acc_list: "+str(acc_list))
        print("smote_time_list: "+str(smote_time_list))
        print("TOTAL smote_time_list: "+str(sum(smote_time_list)))
        print("\n\npruned_weights_list: "+str(pruned_weights_list))
        print("TOTAL pruned paramters: "+str(sum(pruned_weights_list)))
        print("total_nn_weights_list: "+str(total_nn_weights_list))
        print("TOTAL paramters: "+str(sum(total_nn_weights_list)))
        print("pruned_weights_percent_list: "+str(pruned_weights_percent_list))
        print("pruning_time_list: "+str(pruning_time_list))
        print("TOTAL pruning_time_list: "+str(sum(pruning_time_list)))
        


#======================================CLOUD CODE END HERE========================================
















#======================================EDGE CODE DOWN HERE========================================
class EdgeClassifier(nn.Module):
    def __init__(self):
        super(EdgeClassifier, self).__init__()
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons) #2 output nodes initially
        self.drop1 = nn.Dropout(0.25)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop1(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

class EdgeTempClassifier(nn.Module):
    def __init__(self):
        super(EdgeTempClassifier, self).__init__()
        self.drop2 = nn.Dropout(0.25)
        self.softmax_layer = nn.Linear(feature_extractor_size,starting_softmax_neurons)
    
    def forward(self, x):
        x = x.view(-1,feature_extractor_size) #FLATTENING OPERATION 1*1*512 IS OUTPUT AFTER THE PREVIOUS LAYER
        x = self.drop2(x)
        x = self.softmax_layer(x) #LAST LAYER DOES NOT NEED SOFTMAX BECAUSE THE LOSS FUNCTION WILL TAKE CARE OF IT
        return x

def TorchNormalizedTrainingData():
    print("Get sorted training data\n")
    img_list = []
    lab_list = []
    with torch.no_grad():
        for data in Trainloader_fake:
            images, labels = data
            for j in range(0,len(images)):
              img_list.append(np.array(images[j]))
              lab_list.append(int(labels[j]))

    img_list = np.array(img_list)    
    lab_list = np.array(lab_list)
    #Sort training dataset wrt to labels

    sorted_X_train = [] #this will contains training images sorted based on the label number
    sorted_Y_train = [] #label
    start_index = 0
    end_index = 100
    for j in range(start_index,end_index):
      searchval = j
      all_positions_of_label = np.where(lab_list == searchval)[0] #Find all indexes that contains `searchval`
      for j in range(0,len(all_positions_of_label)):
          image = img_list[all_positions_of_label[j]]
          sorted_X_train.append(image)
          sorted_Y_train.append(searchval)
    
    sorted_X_train = np.array(sorted_X_train)
    sorted_X_train = sorted_X_train.reshape(50000, 32, 32, 3)
    #sorted_Y_categorical_train = np_utils.to_categorical(sorted_Y_train,end_index-start_index)

    return sorted_X_train,sorted_Y_train#,sorted_Y_categorical_train

def TorchNormalizedTestingData():
    img_list = []
    lab_list = []
    with torch.no_grad():
      for data in Testloader:
        images, labels = data
        for j in range(0,len(images)):
          img_list.append(np.array(images[j]))
          lab_list.append(int(labels[j]))

    img_list = np.array(img_list)    
    lab_list = np.array(lab_list)
    #Sort training dataset wrt to labels

    sorted_X_test = [] #this will contains training images sorted based on the label number
    sorted_Y_test = [] #label
    start_index = 0
    end_index = 100
    for j in range(start_index,end_index):
      searchval = j
      all_positions_of_label = np.where(lab_list == searchval)[0] #Find all indexes that contains `searchval`
      for j in range(0,len(all_positions_of_label)):
        image = img_list[all_positions_of_label[j]]
        sorted_X_test.append(image)
        sorted_Y_test.append(searchval)
    sorted_X_test = np.array(sorted_X_test)
    sorted_X_test = sorted_X_test.reshape(10000, 32, 32, 3)
    #sorted_Y_categorical_test = np_utils.to_categorical(sorted_Y_test,end_index-start_index)
    
    return sorted_X_test,sorted_Y_test#,sorted_Y_categorical_test

#ANY DATA RECEIVED BY THE CLOUD WILL BE TAKEN CARE OF BY THIS FUNCTION
def ReceiveDataFromCloud(client):
    recevied_data = " "
    received_result = ""
    print("receving start")
    while recevied_data[-1]!="!":
        data = client.recv(10000).decode("utf-8")
        recevied_data += data
        received_result += data
    #print("data received is: "+str(received_result)+"\n")
    #pos = (int)(received_result[0])
    client.close()
    return received_result
    #return 0

#FUNCTION FOR CONNECTED TO GOOGLE CLOUD VM AND TRANSMITTING DATA
def SendDataToCloud(Feature):

    #Connect to the cloud and send the data
    client = socket.socket(socket.AF_INET,socket.SOCK_STREAM)#AF_NET = ipv4, SOCK_STREAM = tcp oriented
    #ip = '34.80.68.8'
    ip = '34.80.253.190'
    #ip = '104.155.209.67'
    port = 3389
    address = (ip,port) #address of the server
    client.connect(address)
    print("sending to the cloud \n")
    
    #Send the string to the cloud
    client.send(Feature.encode())
    print("sent to the cloud \n")

    #Receive response from the server and only then continue
    received_string = ReceiveDataFromCloud(client)
    print("recvd from the cloud \n")
    return received_string

#FUNCTION FOR SENDING TEST DATASET TO THE CLOUD FOR COMPUTING CLASSIFICATION ACCURACY ON TEST DATASET
def CalculateModelAccuracy(squeezenet,device,Testloader,batch_size):
    
    overall_string_tensor = ""
    with torch.no_grad():
        for data in Testloader:
            images, labels = data

            images = images.clone().detach()
            images = images.view(len(images),3,32,32) #RESHAPE THE IMAGES
            labels = labels.long() #MUST CONVERT LABEL TO LONG FORMAT

            images=images.to(device)
            labels=labels.to(device)

            outputs = squeezenet.features(images)

            #Convert the tensor to string
            string_tensor = TensorToString(labels,outputs)
            overall_string_tensor = overall_string_tensor + string_tensor + "k"

        #Send the string to the cloud
        overall_string_tensor = overall_string_tensor[:-1] + "c!"
        Cloud(overall_string_tensor)
    

def GetLosses(Trainloader_trail,device,cls):

    special_criterion = nn.CrossEntropyLoss(reduction='none')

    initial_loss_list = []

    image_loss_dictionary = collections.defaultdict(list)
    label_loss_dictionary = collections.defaultdict(list)

    for _, data in enumerate(Trainloader_trail, 0):
        inputs, labels = data

        inputs = inputs.clone().detach()
        inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
        labels = labels.long()
                
        inputs=inputs.to(device)
        labels=labels.to(device)

        with torch.no_grad():
            outputs = cls.forward(inputs)

            #compute softmax here
            """
            softmax_function = nn.Softmax(dim=1)
            softmax_output = softmax_function(outputs)
            loss = []
            for j in range(len(softmax_output)):
                loss.append(softmax_output[j][labels[j]])
            """
            loss = special_criterion(outputs, labels)
            for j in range(0,len(loss)):
                initial_loss_list.append(loss[j].item())
                image_loss_dictionary[loss[j].item()].append(inputs[j])
                label_loss_dictionary[loss[j].item()].append(labels[j].cpu().numpy())

    return initial_loss_list, image_loss_dictionary, label_loss_dictionary

#FUNCTION FOR COMPUTING THE STATISTICS OF THE SORTED LOSS VALUES OF THE TRAINING IMAGES
def GetLossStatistics(sorted_initial_list):
    #Calculate mean
    mean = statistics.mean(sorted_initial_list)

    #Calculate median
    median = statistics.median(sorted_initial_list)

    #Calculate standard deviation
    standard_deviation = statistics.stdev(sorted_initial_list)

    
    #mid_range_loss = sorted_initial_list[int(len(sorted_initial_list)/2)]
    #C = []
    #for i in range(0,len(sorted_initial_list)):
    #    diff_sqaured = (sorted_initial_list[i]-mid_range_loss)**2
    #    C.append(diff_sqaured)
    #standard_deviation = math.sqrt(sum(C)/len(C))
    
    return mean, median, standard_deviation

def MannWhitneyTest(loss_comparison_list,losses):
    alpha = 0.05

    A = loss_comparison_list
    B = losses
    #C = A + B #Step 1: Concatenate the lists
    C = []
    for i in range(0,len(A)):
      C.append(A[i])
    for i in range(0,len(B)):
      C.append(B[i])
    #C.extend(A)
    #C.extend(B)
    #print("Concatenated: "+str(C))

    ranks = ss.rankdata(C) #Compute the ranks of the concatenrated lists
    #print("ranks: "+str(ranks))

    ranksum_A = sum(ranks[0:len(A)]) #Ranksum of list A
    ranksum_B = sum(ranks[len(A):]) #Ranksum of list B

    #print("ranksum_A: "+str(ranksum_A))
    #print("ranksum_B: "+str(ranksum_B))

    n1 = len(A) #Number of of elements in list A
    n2 = len(B) #Number of of elements in list B

    #print("n1: "+str(n1))
    #print("n2: "+str(n2))

    mean = (n1*(n1 + n2 + 1))/2.0 #Mean
    stddev = math.sqrt((n1*n2*(n1 + n2 + 1))/12.0) #Standard deviation

    #print("mean: "+str(mean))
    #print("stddev: "+str(stddev))

    z = (ranksum_A - mean)/stddev #Z-test
    p_value = norm.cdf(z) #p value

    #print("z: "+str(z))
    #print("p_value: "+str(p_value))

    if(p_value < alpha): #Enough evidence that 2 sets are different
        return 1
    else: #Not enough evidence that 2 sets are different
        return 0

def Kruskal_Wallis_Test(array_list):
    concatenated_list = []
    for i in range(0,len(array_list)):
        concatenated_list += array_list[i]
    ranks = rankdata(concatenated_list)
    n_T = len(concatenated_list)
    R_i = []
    n_i = []
    start = 0
    for i in range(0,len(array_list)):
        end = start + len(array_list[i])
        R_i.append(sum(ranks[start:end]))
        n_i.append(len(array_list[i]))
        start = end

    part_1 = 12/(n_T*(n_T+1))
    part_2 = sum([i / j for i, j in zip([i**2 for i in R_i], n_i)])
    part_3 = 3*(n_T + 1)
    H = (part_1*part_2)-part_3
    #print("H: "+str(H))

    chi_square_value = stats.chi2.ppf(0.95 , len(array_list)-1)
    #print(chi_square_value)
    return H, chi_square_value

#FUNCTION FOR FINDING THE CUTOFF LOSS VALUE AND DISCARDING THE SAMPLES
old_mean_loss = 0
def Cutoff(mean,median,standard_deviation,image_loss_dictionary,label_loss_dictionary,depth,width,height):
    sorted_loss_list = sorted(image_loss_dictionary)
    number_of_samples_to_be_dropped = 0
       
    cutoff_loss_value = min(sorted_loss_list)
    min_loss_value = min(sorted_loss_list)
    for i in range(0,len(sorted_loss_list)):
        distance = median - sorted_loss_list[i]
        if(distance < standard_deviation):
            cutoff_loss_value = sorted_loss_list[i]
            print("min loss: "+str(min_loss_value)+", distance: "+str(distance)+" , std: "+str(standard_deviation))
            break
    
    #cutoff_loss_value = cutoff_loss_value - (0.25*(cutoff_loss_value-min_loss_value))
    number_of_samples_to_be_dropped = 0
    for i in range(0,len(sorted_loss_list)):
        current_loss = sorted_loss_list[i]
        if(current_loss < cutoff_loss_value):
            number_of_samples_to_be_dropped += 1
        else:
            break
    
    omega = 1.0
    number_of_samples_to_be_dropped  = int(omega * number_of_samples_to_be_dropped)
    

    """   
    #WRSTS sampling
    base_losses = sorted_loss_list
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        val = MannWhitneyTest(base_losses,test_losses)
        if(val == 1):
            print("mann-whitney deviation is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    """

    """
    #MTS sampling
    base_losses = sorted_loss_list
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        stat, p, med, tbl = median_test(base_losses,test_losses)
        if(p < 0.05):
            print("median deviation is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    """

    """
    #KWT Sampling
    base_losses = sorted_loss_list
    for cnt in range(0,len(sorted_loss_list)):
        test_losses = sorted_loss_list[cnt:]
        array_list = [test_losses,base_losses]
        #array_list.append(test_losses)
        #array_list.append(base_losses)
        #H,p=stats.mstats.kruskalwallis(array_list)
        H,chi_square_value = Kruskal_Wallis_Test(array_list)
        #print("HH: "+str(H))
        #print("chi_square_value: "+str(chi_square_value))
        #chi_square_value = stats.chi2.ppf(0.95 , len(array_list)-1)
        if(H > chi_square_value): #If true then reject the null hypothesis which means the 2 groups not the same
            print("Kruskal wallis is evident!, number of samples to be discarded: "+str(cnt))
            number_of_samples_to_be_dropped = cnt
            break
    """
    print("number_of_samples_to_be_dropped: "+str(number_of_samples_to_be_dropped))

    #get all the images and labels above the cutoff value
    number_of_samples_to_selected = 0
    new_images = 0
    new_labels = []
    filtered_original_losses = []

    #"""
    number_of_samples_to_be_selected = 0
    #randomly select losses
    if(number_of_samples_to_be_dropped > 0):
        number_of_samples_to_be_selected = len(sorted_loss_list) - number_of_samples_to_be_dropped
        print("number_of_samples_to_be_selected: "+str(number_of_samples_to_be_selected))
        randomly_selected_losses = random.sample(sorted_loss_list, k=number_of_samples_to_be_selected)
    else:
        randomly_selected_losses = random.sample(sorted_loss_list, k=len(sorted_loss_list))
        print("number_of_samples_to_be_selected: "+str(len(sorted_loss_list)))
        print("must select all the samples now!!!!! ")
    
    
    #RANDOM SAMPLING
    for i in range(0,len(randomly_selected_losses)):
        key_loss = randomly_selected_losses[i]
        #if(key_loss not in randomly_selected_losses):
        selected_image = image_loss_dictionary[key_loss]
        selected_label = label_loss_dictionary[key_loss]

        #if(len(selected_image)>1):
        #      print("length of selected_image: "+str(len(selected_image)))
        if(number_of_samples_to_selected == 0):
            for j in range(0,len(selected_image)): # aloss key can have duplicates
                if(j==0):
                    new_images = selected_image[j]
                else:
                    new_images = torch.cat([new_images,selected_image[j]],dim=0)
                #print("type of label is: "+str(type(selected_label[j])))
                new_labels.append(selected_label[j])
                filtered_original_losses.append(key_loss)
        else:
            #print("betrne")
            for j in range(0,len(selected_image)): # aloss key can have duplicates
                new_images = torch.cat([new_images,selected_image[j]],dim=0)
                new_labels.append(selected_label[j])
                filtered_original_losses.append(key_loss)
        number_of_samples_to_selected += len(selected_image)
    """
    #ACTIVE LEARNING ENTROPY BASED SAMPLING
    cutoff_loss_value = sorted_loss_list[number_of_samples_to_be_dropped]
    for i in range(len(sorted_loss_list)-1,-1,-1):
        #print("i is: "+str(i))
        key_loss = sorted_loss_list[i]
        #print("key loss type is: "+str(type(key_loss))+" and value is: "+str(key_loss))
        if(key_loss>=cutoff_loss_value):
            selected_image = image_loss_dictionary[key_loss]
            selected_label = label_loss_dictionary[key_loss]
            #print("key: "+str(key_loss)+" and value: "+str(data[key_loss]))
            #if(len(selected_image)>1):
              #print("length of selected_image: "+str(len(selected_image)))
            if(number_of_samples_to_selected == 0):
                for j in range(0,len(selected_image)): # aloss key can have duplicates
                    if(j==0):
                        new_images = selected_image[j]
                    else:
                        new_images = torch.cat([new_images,selected_image[j]],dim=0)
                    #print("type of label is: "+str(type(selected_label[j])))
                    new_labels.append(selected_label[j])
                    filtered_original_losses.append(key_loss)
            else:
                #print("betrne")
                for j in range(0,len(selected_image)): # aloss key can have duplicates
                    new_images = torch.cat([new_images,selected_image[j]],dim=0)
                    new_labels.append(selected_label[j])
                    filtered_original_losses.append(key_loss)
            number_of_samples_to_selected += len(selected_image)
        else:
            break
    """
    new_images = new_images.reshape((number_of_samples_to_selected,depth,width,height))
    print("new_images shape: "+str(new_images.shape))

    new_labels = torch.from_numpy(np.asarray(new_labels)).float()
    print("new_labels shape: "+str(new_labels.shape))

    #filtered_original_losses = torch.from_numpy(np.asarray(filtered_original_losses)).float()
    #print("filtered_original_losses shape: "+str(filtered_original_losses.shape))
    return new_images,new_labels#,filtered_original_losses

#MAIN FUNCTION FOR PERFORMING DATA SAMPLING
def GetusefulData(images_train,labels_train,batch_size,cls):

    traindataset = TensorDataset( images_train, Tensor(labels_train) )
    Trainloader_trail = DataLoader(traindataset, batch_size= batch_size, shuffle=False)

    #Do single forward pass and get the losses
    #cl.eval()
    initial_loss_list, image_loss_dictionary, label_loss_dictionary = GetLosses(Trainloader_trail,device,cls)
    sorted_initial_list = sorted(initial_loss_list)
    mean, median, standard_deviation = GetLossStatistics(sorted_initial_list)
    #print("mean: "+str(mean))
    #print("median: "+str(median))
    #print("standard_deviation: "+str(standard_deviation))
    new_images, new_labels = Cutoff(mean,median,standard_deviation,image_loss_dictionary,label_loss_dictionary,feature_d,feature_w,feature_h)
    

    rejection_rate = (1.0 - (len(new_images)/len(images_train)))*100.0
    print("rejection rate: "+str(rejection_rate))
    return new_images.cpu().numpy(), new_labels.cpu().numpy()



#FUNCTION TO CONVERT TENSOR TO STRING
def TensorToString(labels,outputs):

    tensor = outputs
    tensor = tensor.view(-1,outputs.shape[1]*outputs.shape[2]*outputs.shape[3]) #flatten the tensor

    string_tensor = ""
    string_tensor = str(outputs.shape[1]) + "," + str(outputs.shape[2]) + "," + str(outputs.shape[3]) + ","
    mini_batch_length = len(labels)
    for i in range(0,mini_batch_length):
        string_tensor = string_tensor + str(labels[i].item()) + ","


    new_string = string_tensor
    for i in range(0,mini_batch_length):
        numpy_tensor = tensor[i].cpu().numpy()
        out_string = ','.join(str(x) for x in numpy_tensor)
        new_string = new_string + out_string + ","
    new_string = new_string[:-1] #remove last comma
    #new_string = new_string + "t!" 

    string_tensor = new_string
    return string_tensor

#FUNCTION TO DECODE WEIGHTS AND BIASES (SEPARATING WEIGHTS & BIASES FROM COMMAS)
def DecodeWeightsAndBiases(message):
    task_character = message[-2]

    bias_string = message[message.find("b")+2:message.find(task_character)-1]
    bias_list = list(map(float, bias_string.split(",")))

    weight_string = message[0:message.find("i")-1]
    weight_list = list(map(float, weight_string.split(","))) 

    indices_string = message[message.find("i")+2:message.find("b")-1]
    indices_list = indices_string.split(",")

    return weight_list, bias_list, indices_list, task_character

def DataSampling(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl, edge_temp_cl,number_of_classes_to_be_added_ever_increment,incremental_i):

    new_images = 0
    print("per class data sampling begins here \n")
    per_class_labels_list_np_array = np.asarray(per_class_labels_list)
    if(incremental_i == 0):
        edge_cl.eval()
        new_images, new_labels = GetusefulData(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl)
        edge_cl.train()
    else:
        for kk in range(0,len(per_class_labels_list)):
            per_class_labels_list[kk] -= (number_of_classes_to_be_added_ever_increment*incremental_i)
        per_class_labels_list_np_array = np.asarray(per_class_labels_list)
        edge_temp_cl.eval()
        new_images, new_labels = GetusefulData(per_class_features, per_class_labels_list_np_array, Batch_size, edge_temp_cl)
        edge_temp_cl.train()
        for kk in range(0,len(new_labels)):
            new_labels[kk] += (number_of_classes_to_be_added_ever_increment*incremental_i)
    
    return new_images, new_labels

#==========MAIN PROGRAM==========
squeezenet = models.mobilenet_v2(pretrained=True)
squeezenet = squeezenet.to(device)
squeezenet.eval()

edge_cl = EdgeClassifier()
edge_cl = edge_cl.to(device) #MAP THE MODEL ONTO THE GPU
edge_cl.eval()

edge_temp_cl = EdgeTempClassifier()
edge_temp_cl = edge_temp_cl.to(device) #MAP THE MODEL ONTO THE GPU
edge_temp_cl.eval()

edge_criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
edge_optimizer = optim.Adam(edge_cl.parameters(), lr=1e-4) #ADAM OPTIMIZER


training_image_list = 0
training_image_label_list = []
testing_image_list = 0
testing_image_label_list = []
#must use the absolute path
directory = '/home/swarajdubey611/Stanford_dogs'#r"C:\Users\Swaraj\AppData\Local\Programs\Python\Python36\Journal_paper_codes\SqueezeNet\CUB_200_2011\images"
#directory = '/content/gdrive/My Drive/CUB_200_resized'
index_counter = 0
train_counter = 0
test_counter = 0
class_counter = -1
Exemplar_features = 0
Exemplar_labels = 0
classes_per_training_round_count = 0
edge_incremental_i = 0
number_of_classes_to_be_added_ever_increment = starting_softmax_neurons
total_nerons_in_final_layer= starting_softmax_neurons

#Variables for SMOTE
useful_images_list = []
useful_labels_list = []
length_of_dataset_after_filtration = []

Batch_size = 32
criterion = nn.CrossEntropyLoss() #CROSS ENTROPY LOSS FUNCTION
learning_rate = 1e-4
optimizer = optim.Adam(edge_cl.parameters(), lr=learning_rate) #ADAM OPTIMIZER
total_epochs = 15
batch_size = 128 #this batch size is only for data loading

training_features_list = 0
testing_features_list = 0
training_features_list_labels = []
testing_features_list_labels = []
folder_count = 0
train_batch_counter = 0
test_batch_counter = 0
all_testing_features_so_far = 0
all_testing_labels_so_far = []
edge_processing_time_list = []

samples_transmitted_count = 0
length_so_far = 0
number_of_overloads = 0
datasize_sent_list = []
overall_string_tensor = ""
testing_overall_string_tensor = ""
total_append_operation_time = 0

new_images_length_counter = 0
new_images_length_list = []

train_mat = loadmat('/home/swarajdubey611/train_list.mat')
test_mat = loadmat('/home/swarajdubey611/test_list.mat')

training_files = train_mat['file_list']
testing_files = test_mat['file_list']

#total_classes = 200
total_classes = 120
#number_of_increments = int(total_classes/starting_softmax_neurons) #total number of classes to be learnt
#number_of_increments = 2
img_dimension = 85
overall_tic = time.time()
for subdir in os.listdir(directory):
    #print("edge_incremental_i: "+str(edge_incremental_i)+", number_of_increments: "+str(number_of_increments))
    if(edge_incremental_i < total_classes):# (classes_per_training_round_count < number_of_classes_to_be_added_ever_increment) and (incremental_i < number_of_increments)
        print(subdir)
        subdir_absolute_path = directory + '/' + subdir
        class_counter += 1
        folder_count += 1
        
        for f in os.listdir(subdir_absolute_path):
            file_name = subdir + '/' + f
            ori_img = Image.open(os.path.join(subdir_absolute_path,f)) #this is the JPEG image
            width, height = ori_img.size  
            #print("img size: "+str(ori_img.size))
            newsize = (img_dimension, img_dimension) 
            ori_img = ori_img.resize(newsize) 
            img = transform(ori_img)
            if(img.shape[0] == 3):#accept only RGB images
                img = img.reshape((1,3,img_dimension,img_dimension)) #reshape the image tensor
                
                #Extract the image feature here itself to prevents a burden on the IoT device
                #img = torch.FloatTensor(img)
                #img = img.view(1,3,img_dimension,img_dimension)
                #feature = squeezenet(img)
                #print("file name: "+str(file_name))
                try: #this means the file is in the training set
                    result = np.where(training_files == file_name)[0][0]
                    #print("inside the training dataset")
                    if(train_counter == 0): #first time appending
                        training_image_list = img
                        #training_image_list = feature
                        train_counter = 1
                    else:
                        training_image_list = np.concatenate((training_image_list,img),axis=0)
                        #training_image_list = torch.cat([training_image_list,feature],dim=0)
                    training_image_label_list.append(class_counter)

                except: #this means the file is in the testing set
                    #print("inside the testing dataset")
                    if(test_counter == 0): #first time appending
                        testing_image_list = img
                        #testing_image_list = feature
                        test_counter = 1
                    else:
                        testing_image_list = np.concatenate((testing_image_list,img),axis=0)
                        #testing_image_list = torch.cat([testing_image_list,feature],dim=0)
                    testing_image_label_list.append(class_counter)
            
            
        if(folder_count == 1): #Send append message to the cloud
            
            print("Training images size: "+str(len(training_image_label_list)))
            print("Testing images size: "+str(len(testing_image_label_list)))
            print("Starting data sampling")
            folder_count = 0
            edge_tic = time.time()
            #OBTAIN FEATURES OF ALL SAMPLES IN THE SUB FOLDER
            trail_train_dataset = TensorDataset( Tensor(training_image_list), Tensor(np.asarray(training_image_label_list)) )
            Trialtrainloader = DataLoader(trail_train_dataset, batch_size= batch_size, shuffle=False)
            training_image_list = 0
            training_image_label_list = []

            per_class_features = 0
            per_class_features_counter = 0
            per_class_labels_list = []
            for i, data in enumerate(Trialtrainloader, 0):
                inputs, _ = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),3,img_dimension,img_dimension) 
                inputs=inputs.to(device)

                for _ in range(len(inputs)):
                    per_class_labels_list.append(class_counter)

                with torch.no_grad():
                    outputs = squeezenet.features(inputs)
                    print("outputs size: "+str(outputs.shape))
                    if(per_class_features_counter==0):
                        per_class_features = outputs
                        per_class_features_counter = 1
                    else:
                        per_class_features = torch.cat([per_class_features,outputs],dim=0)


            #Carry out data sampling here
            per_class_labels_list_np_array = np.asarray(per_class_labels_list)
            new_images, new_labels = DataSampling(per_class_features, per_class_labels_list_np_array, Batch_size, edge_cl, edge_temp_cl,number_of_classes_to_be_added_ever_increment,edge_incremental_i)
            
            new_images_length_counter += len(new_images)
            print("per class data sampling ends here \n")
            Trainloader_iterations = 0
            if(len(new_images) % batch_size == 0):
                Trainloader_iterations = len(new_images) // batch_size
            else:
                Trainloader_iterations = (len(new_images) // batch_size) + 1
            edge_toc = time.time()
            edge_processing_time_list.append(edge_toc-edge_tic)

            samples_transmitted_count += len(new_images)
            
            #Convert the filtered data samples to string format
            traindataset_new = TensorDataset( Tensor(new_images), Tensor(new_labels))
            Trainloader = DataLoader(traindataset_new, batch_size= batch_size, shuffle=True)

            ti = 0
            for i, data in enumerate(Trainloader, 0): #This will only run once i.e. only 1 epoch        
                mini_batch_tic = time.time()
                inputs, labels = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),feature_d,feature_w,feature_h)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                outputs = inputs

                #Send the data to the cloud device (images and the labels)
                string_tensor = TensorToString(labels,outputs)
                overall_string_tensor = overall_string_tensor + string_tensor + "k"
                mini_batch_toc = time.time()
                time_diff = mini_batch_toc - mini_batch_tic

                if(i==0):
                    ti = time_diff
                else:
                    if((time_diff - ti) > 0.5):
                        print("RAM overload!")
                        if(i == Trainloader_iterations - 1): #RAM overload at the last iteration! so start training already
                            overall_string_tensor = overall_string_tensor[:-1] + "t!"
                            append_ticc = time.time()
                            Cloud(overall_string_tensor)
                            append_tocc = time.time()
                            total_append_operation_time += (append_tocc-append_ticc)
                        else:
                            overall_string_tensor = overall_string_tensor[:-1] + "a!" #remove the last k and add a! at the end
                            append_ticc = time.time()
                            Cloud(overall_string_tensor)
                            append_tocc = time.time()
                            total_append_operation_time += (append_tocc-append_ticc)
                        length_so_far += len(overall_string_tensor)
                        length_so_far *= 1e-6
                        datasize_sent_list.append(length_so_far)
                        overall_string_tensor = ""
                        number_of_overloads += 1


            #get testing features of images and only save that
            ti = 0
            trail_test_dataset = TensorDataset( Tensor(testing_image_list), Tensor(np.asarray(testing_image_label_list)) )
            Trialtestloader = DataLoader(trail_test_dataset, batch_size= batch_size, shuffle=False)
            testing_image_list = 0
            testing_image_label_list = []
            for i, data in enumerate(Trialtestloader, 0):
                mini_batch_tic = time.time()

                inputs, labels = data

                inputs = inputs.clone().detach()
                inputs = inputs.view(len(inputs),3,img_dimension,img_dimension)
                labels = labels.long()

                inputs=inputs.to(device)
                labels=labels.to(device)

                with torch.no_grad():
                    outputs = squeezenet.features(inputs)
                string_tensor = TensorToString(labels,outputs)
                testing_overall_string_tensor = testing_overall_string_tensor + string_tensor + "k"
                mini_batch_toc = time.time()
                time_diff = mini_batch_toc - mini_batch_tic
                if(i==0):
                    ti = time_diff
                else:
                    if((time_diff - ti) > 0.5):
                        print("RAM overload!")
                        if(classes_per_training_round_count == number_of_classes_to_be_added_ever_increment-1):
                            testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "c!"
                            Cloud(testing_overall_string_tensor)
                        else:
                            testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "h!"
                            Cloud(testing_overall_string_tensor)
                        testing_overall_string_tensor = ""

            #reset all the variables
            training_image_list = 0
            testing_image_list = 0
            train_counter = 0
            test_counter = 0
            training_image_label_list = []
            testing_image_label_list = []


        if(classes_per_training_round_count == number_of_classes_to_be_added_ever_increment-1): #Send to the cloud for training

            new_images_length_list.append(samples_transmitted_count)
            samples_transmitted_count = 0
            print("before training edge_incremental_i: "+str(edge_incremental_i))
            if(len(overall_string_tensor)>1):
                overall_string_tensor = overall_string_tensor[:-1] + "t!" #remove the last k and add t! at the end
                length_so_far += len(overall_string_tensor)
                length_so_far *= 1e-6
                datasize_sent_list.append(length_so_far)
                Cloud(overall_string_tensor)        
            overall_string_tensor = ""            


            #Start the classification process
            if(len(testing_overall_string_tensor)>1):
                testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "c!"
                Cloud(testing_overall_string_tensor)
            testing_overall_string_tensor = ""

            #Modify the edge_cl
            print("modify softmax layer edge "+str(edge_cl))
            #Add neurons to the SoftMax classification layer
            
            current_weights = edge_cl.softmax_layer.weight.data
            current_bias = edge_cl.softmax_layer.bias.data

            new_total_nerons_in_final_layer = total_nerons_in_final_layer + number_of_classes_to_be_added_ever_increment
            total_nerons_in_final_layer += number_of_classes_to_be_added_ever_increment
            edge_cl.softmax_layer = nn.Linear(feature_extractor_size,total_nerons_in_final_layer)

            new_weights = -0.0001 * np.random.random_sample((number_of_classes_to_be_added_ever_increment, feature_extractor_size)) + 0.0001 #RANDOMLY INITIALIZED WEIGHTS FOR THE NEW TASKS
            new_weights = torch.from_numpy(new_weights).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
            new_weights = new_weights.to(device)
            new_weights = torch.cat([current_weights,new_weights],dim=0) #Concatenate the new initialized weights with the old weights for old task
            edge_cl.softmax_layer.weight.data.copy_(torch.tensor(new_weights, requires_grad=True))

            new_biases = np.zeros(number_of_classes_to_be_added_ever_increment)
            new_biases = torch.from_numpy(new_biases).float() #CONVERT NEWLY RANDONLY INITIALIZED WEIGHTS TO A TORCH TENSOR
            new_biases = new_biases.to(device)
            new_biases = torch.cat([current_bias,new_biases],dim=0)
            edge_cl.softmax_layer.bias.data.copy_(torch.tensor(new_biases, requires_grad=True))

            #After adding neurons to the final layer, the network needs to be pushed to the GPU again
            edge_cl = edge_cl.to(device)


            #Assign random weights to tempcl
            edge_modifiable_cl = copy.deepcopy(edge_cl)
            edge_modifiable_cl_weight = edge_modifiable_cl.softmax_layer.weight[(total_nerons_in_final_layer - number_of_classes_to_be_added_ever_increment):total_nerons_in_final_layer] #base weights
            edge_modifiable_cl_bias = edge_modifiable_cl.softmax_layer.bias[(total_nerons_in_final_layer - number_of_classes_to_be_added_ever_increment):total_nerons_in_final_layer] #base biases
            edge_temp_cl.softmax_layer.weight.data.copy_(torch.tensor(edge_modifiable_cl_weight, requires_grad=True))
            edge_temp_cl.softmax_layer.bias.data.copy_(torch.tensor(edge_modifiable_cl_bias, requires_grad=True))

            #reset all the variables for the next incremental training round
            training_features_list = 0
            training_features_list_labels = []
            testing_features_list = 0
            testing_features_list_labels = []
            train_batch_counter = 0
            test_batch_counter = 0
            edge_incremental_i += 1
            classes_per_training_round_count = -1

            print("after training edge_incremental_i: "+str(edge_incremental_i))
        else: #Send to the cloud as append append
            print("acloud append")
            if(len(overall_string_tensor)>1):
                overall_string_tensor = overall_string_tensor[:-1] + "a!" #remove the last k and add a! at the end
                length_so_far += len(overall_string_tensor)
                length_so_far *= 1e-6
                datasize_sent_list.append(length_so_far)
                append_ticc = time.time()
                Cloud(overall_string_tensor)
                append_tocc = time.time()
                total_append_operation_time += (append_tocc-append_ticc)
            overall_string_tensor = ""

            if(len(testing_overall_string_tensor)>1):
                testing_overall_string_tensor = testing_overall_string_tensor[:-1] + "h!"
                Cloud(testing_overall_string_tensor)
            testing_overall_string_tensor = ""

    classes_per_training_round_count += 1

stop_cloud_message = "d!"
Cloud(stop_cloud_message)
overall_toc = time.time()
print("=========================EDGE RESULTS===========================")
print("overall time: "+str(overall_toc - overall_tic))
print("new_images_length_list: "+str(new_images_length_list))
print("total images sent: "+str(sum(new_images_length_list)))
print("amount of data sent so far (MB): ")
print(datasize_sent_list)
print("TOTAL data sent (MB): ")
print(sum(datasize_sent_list))
